"""PyTorch Calibrated training utility functions."""
import time
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import torch
import torchmetrics
from pydantic import BaseModel
from tqdm import trange

from .constants import MISSING_CATEGORY_VALUE, MISSING_NUMERICAL_VALUE
from .data import CSVData
from .enums import FeatureType, LossType, Metric
from .features import CategoricalFeature, NumericalFeature
from .models import CalibratedLinear
from .types import (
    CategoricalFeatureConfig,
    Dataset,
    FeatureAnalysis,
    LinearConfig,
    NumericalFeatureConfig,
    PipelineConfig,
    TrainingConfig,
    TrainingResults,
)


class PerEpochResults(BaseModel):
    """Container for the per-epoch results of training a PyTorch Calibrated model."""

    train_loss_by_epoch: List[float]
    train_primary_metric_by_epoch: List[float]
    val_loss_by_epoch: List[float]
    val_primary_metric_by_epoch: List[float]


def create_features(
    feature_configs: Dict[str, Union[CategoricalFeatureConfig, NumericalFeatureConfig]],
    train_csv_data: CSVData,
) -> List[Union[CategoricalFeature, NumericalFeature]]:
    """Returns a list of PyTorch Calibrated feature configs."""
    features = []

    for feature_name, feature_config in feature_configs.items():
        if feature_config.type == FeatureType.CATEGORICAL:
            features.append(
                CategoricalFeature(
                    feature_name=feature_name,
                    categories=feature_config.categories,
                    missing_input_value=MISSING_NUMERICAL_VALUE,
                )
            )
        else:  # FeatureType.NUMERICAL
            features.append(
                NumericalFeature(
                    feature_name=feature_name,
                    data=train_csv_data(feature_name),
                    num_keypoints=feature_config.num_keypoints,
                    input_keypoints_init=feature_config.input_keypoints_init,
                    missing_input_value=MISSING_NUMERICAL_VALUE,
                    monotonicity=feature_config.monotonicity,
                    projection_iterations=feature_config.projection_iterations,
                )
            )

    return features


def create_loss(loss_type: LossType) -> torch.nn.Module:
    """Returns a Torch loss function from the given `LossType`."""
    if loss_type == LossType.BINARY_CROSSENTROPY:
        return torch.nn.BCEWithLogitsLoss()
    if loss_type == LossType.HINGE:
        return torch.nn.HingeEmbeddingLoss()
    if loss_type == LossType.HUBER:
        return torch.nn.HuberLoss()
    if loss_type == LossType.MAE:
        return torch.nn.L1Loss()
    if loss_type == LossType.MSE:
        return torch.nn.MSELoss()

    raise ValueError(f"Unknown loss type: {loss_type}")


def create_metric(metric: Metric) -> torchmetrics.Metric:
    """Returns a torchmetric Metric for the given `Metric`."""
    if metric == Metric.AUC:
        return torchmetrics.AUROC("binary")
    if metric == Metric.MAE:
        return torchmetrics.MeanAbsoluteError()
    if metric == Metric.MSE:
        return torchmetrics.MeanSquaredError()

    raise ValueError(f"Unknown metric: {metric}")


def create_model(
    features: List[Union[CategoricalFeature, NumericalFeature]],
    model_config: LinearConfig,
):
    """Returns a PTCM model config constructed from the given `ModelConfig`."""
    return CalibratedLinear(
        features,
        output_min=model_config.output_min,
        output_max=model_config.output_max,
        use_bias=model_config.use_bias,
        output_calibration_num_keypoints=None
        if not model_config.output_calibration
        else model_config.output_calibration_num_keypoints,
    )


def train_model(  # pylint: disable=too-many-locals
    target: str,
    primary_metric: Metric,
    train_csv_data: CSVData,
    val_csv_data: CSVData,
    pipeline_config: PipelineConfig,
    model_config: LinearConfig,
    training_config: TrainingConfig,
) -> Tuple[CalibratedLinear, PerEpochResults, torch.nn.Module, torchmetrics.Metric]:
    """Trains a PyTorch Calibrated model according to the given config."""
    features = create_features(pipeline_config.feature_configs, train_csv_data)
    model = create_model(features, model_config)

    optimizer = torch.optim.Adam(
        model.parameters(recurse=True), training_config.learning_rate
    )
    loss_fn = create_loss(training_config.loss_type)
    metric_fn = create_metric(primary_metric)

    train_loss_by_epoch = []
    train_primary_metric_by_epoch = []
    val_loss_by_epoch = []
    val_primary_metric_by_epoch = []
    train_csv_data.prepare(features, target)
    val_csv_data.prepare(features, target)
    val_examples, val_targets = list(val_csv_data.batch(val_csv_data.num_examples))[0]
    for _ in trange(training_config.epochs, desc="Training Progress"):
        train_prediction_tensors = []
        train_target_tensors = []
        for example_batch, target_batch in train_csv_data.batch(
            training_config.batch_size
        ):
            optimizer.zero_grad()
            outputs = model(example_batch)
            train_prediction_tensors.append(outputs)
            train_target_tensors.append(target_batch)
            loss = loss_fn(outputs, target_batch)
            loss.backward()
            optimizer.step()
            model.constrain()
        with torch.no_grad():
            predictions = torch.cat(train_prediction_tensors)
            targets = torch.cat(train_target_tensors)
            train_loss = loss_fn(predictions, targets)
            train_loss_by_epoch.append(train_loss.tolist())
            train_metric = metric_fn(predictions, targets)
            train_primary_metric_by_epoch.append(train_metric.tolist())
            val_predictions = model(val_examples)
            val_loss = loss_fn(val_predictions, val_targets)
            val_loss_by_epoch.append(val_loss.tolist())
            val_metric = metric_fn(val_predictions, val_targets)
            val_primary_metric_by_epoch.append(val_metric.tolist())

    per_epoch_results = PerEpochResults(
        train_loss_by_epoch=train_loss_by_epoch,
        train_primary_metric_by_epoch=train_primary_metric_by_epoch,
        val_loss_by_epoch=val_loss_by_epoch,
        val_primary_metric_by_epoch=val_primary_metric_by_epoch,
    )

    return model, per_epoch_results, loss_fn, metric_fn


def extract_feature_analyses(
    model: CalibratedLinear,
    feature_configs: Dict[str, Union[CategoricalFeatureConfig, NumericalFeatureConfig]],
    data: pd.DataFrame,
) -> Dict[str, FeatureAnalysis]:
    """Extracts feature statistics and calibration weights for each feature.

    Args:
        model: A (pytorch) calibrated model.
        feature_configs: A mapping from feature name to feature config.
        data: The training + validation data for this model.

    Returns:
        A dictionary mapping feature name to `FeatureAnalysis` instance.
    """
    feature_analyses = {}

    for feature_name, calibrator in model.calibrators.items():
        feature_config = feature_configs[feature_name]

        if feature_config.type == FeatureType.NUMERICAL:
            keypoints_inputs_numerical = [
                float(x) for x in calibrator.keypoints_inputs()
            ]
            keypoints_inputs_categorical = []
        else:
            keypoints_inputs_numerical = []
            keypoints_inputs = feature_config.categories + [MISSING_CATEGORY_VALUE]
            keypoints_inputs_categorical = [str(ki) for ki in keypoints_inputs]

        fa_dict = {
            "feature_name": feature_name,
            "feature_type": feature_config.type,
            "keypoints_inputs_numerical": keypoints_inputs_numerical,
            "keypoints_inputs_categorical": keypoints_inputs_categorical,
            "keypoints_outputs": [float(x) for x in calibrator.keypoints_outputs()],
        }

        if feature_config.type == FeatureType.NUMERICAL:
            fa_dict.update(
                {
                    "min": float(np.min(data[feature_name].values)),
                    "max": float(np.max(data[feature_name].values)),
                    "mean": float(np.mean(data[feature_name].values)),
                    "median": float(np.median(data[feature_name].values)),
                    "std": float(np.std(data[feature_name].values)),
                }
            )

        feature_analyses[feature_name] = FeatureAnalysis(**fa_dict)

    return feature_analyses


def extract_linear_coefficients(
    linear_model: CalibratedLinear,
    features: List[str],
) -> Dict[str, float]:
    """Extracts linear coefficients from a PyTorch `CalibratedLinear` model."""
    linear_coefficients = dict(
        zip(features, linear_model.linear.kernel.detach().numpy().flatten())
    )
    if linear_model.use_bias:
        linear_coefficients["bias"] = linear_model.linear.bias.detach().numpy()[0]

    return linear_coefficients


def train_and_evaluate_model(  # pylint: disable=too-many-locals
    dataset: Dataset,
    target: str,
    primary_metric: Metric,
    pipeline_config: PipelineConfig,
    model_config: LinearConfig,
    training_config: TrainingConfig,
) -> Tuple[CalibratedLinear, TrainingResults]:
    """Trains a PyTorch Calibrated model according to the given config."""
    train_csv_data = CSVData(dataset.prepared_data.train)
    val_csv_data = CSVData(dataset.prepared_data.val)

    training_start_time = time.time()
    trained_model, per_epoch_results, loss_fn, metric_fn = train_model(
        target,
        primary_metric,
        train_csv_data,
        val_csv_data,
        pipeline_config,
        model_config,
        training_config,
    )
    training_time = time.time() - training_start_time

    evaluation_start_time = time.time()
    test_csv_data = CSVData(dataset.prepared_data.test)
    test_csv_data.prepare(trained_model.features, target)
    x_test, y_test = list(test_csv_data.batch(test_csv_data.num_examples))[0]
    with torch.no_grad():
        evaluation_predictions = trained_model(x_test)
        evaluation_loss = loss_fn(evaluation_predictions, y_test)
        evaluation_metric = metric_fn(evaluation_predictions, y_test)
    evaluation_time = time.time() - evaluation_start_time

    feature_analyses = extract_feature_analyses(
        trained_model,
        pipeline_config.feature_configs,
        pd.concat([train_csv_data.prepared_data, val_csv_data.prepared_data]),
    )

    linear_coefficients = extract_linear_coefficients(
        trained_model, list(pipeline_config.feature_configs.keys())
    )

    training_results = TrainingResults(
        training_time=training_time,
        train_loss_by_epoch=per_epoch_results.train_loss_by_epoch,
        train_primary_metric_by_epoch=per_epoch_results.train_primary_metric_by_epoch,
        val_loss_by_epoch=per_epoch_results.val_loss_by_epoch,
        val_primary_metric_by_epoch=per_epoch_results.val_primary_metric_by_epoch,
        evaluation_time=evaluation_time,
        test_loss=evaluation_loss.tolist(),
        test_primary_metric=evaluation_metric.tolist(),
        feature_analyses=feature_analyses,
        linear_coefficients=linear_coefficients,
    )

    return trained_model, training_results



"""Categorical calibration module.

PyTorch implementation of the categorical calibration module. This module takes in a
single-dimensional input of categories represented as indices and transforms it by
mapping a given category to its learned output value.
"""
from collections import defaultdict
from graphlib import CycleError, TopologicalSorter
from typing import List, Optional, Tuple

import torch

from ..enums import CategoricalCalibratorInit


# pylint: disable-next=too-many-instance-attributes
class CategoricalCalibrator(torch.nn.Module):
    """A categorical calibrator.

    This module takes an input of shape `(batch_size, 1)` and calibrates it by mapping a
    given category to its learned output value. The output will have the same shape as
    the input.

    Attributes:
        - All `__init__` arguments.
        kernel: `torch.nn.Parameter` that stores the categorical mapping weights.

    Example:
    ```python
    inputs = torch.tensor(...)  # shape: (batch_size, 1)
    calibrator = CategoricalCalibrator(
        num_categories=5,
        missing_input_value=-1,
        output_min=0.0
        output_max=1.0,
        monotonicity_pairs=[(0, 1), (1, 2)],
        kernel_init=CateegoricalCalibratorInit.UNIFORM,
    )
    outputs = calibrator(inputs)
    ```
    """

    # pylint: disable-next=too-many-branches
    def __init__(  # pylint: disable=too-many-arguments
        self,
        num_categories: int,
        missing_input_value: Optional[float] = None,
        output_min: Optional[float] = None,
        output_max: Optional[float] = None,
        monotonicity_pairs: Optional[List[Tuple[int, int]]] = None,
        kernel_init: CategoricalCalibratorInit = CategoricalCalibratorInit.UNIFORM,
    ) -> None:
        """Initializes an instance of `CategoricalCalibrator`.

        Args:
            num_categories: The number of known categories.
            missing_input_value: If provided, the calibrator will learn to map all
                instances of this missing input value to a learned output value just
                the same as it does for known categories. Note that `num_categories`
                will be one greater to include this missing category.
            output_min: Minimum output value. If `None`, the minimum output value will
                be unbounded.
            output_max: Maximum output value. If `None`, the maximum output value will
                be unbounded.
            monotonicity_pairs: List of pairs of indices `(i,j)` indicating that the
                calibrator output for index `j` should be greater than or equal to that
                of index `i`.
            kernel_init: Initialization scheme to use for the kernel.

        Raises:
            ValueError: If `monotonicity_pairs` is cyclic.
            ValueError: If `kernel_init` is invalid.
        """
        super().__init__()

        self.num_categories = (
            num_categories + 1 if missing_input_value is not None else num_categories
        )
        self.missing_input_value = missing_input_value
        self.output_min = output_min
        self.output_max = output_max
        self.monotonicity_pairs = monotonicity_pairs
        if monotonicity_pairs:
            self._monotonicity_graph = defaultdict(list)
            self._reverse_monotonicity_graph = defaultdict(list)
            for i, j in monotonicity_pairs:
                self._monotonicity_graph[i].append(j)
                self._reverse_monotonicity_graph[j].append(i)
            try:
                self._monotonically_sorted_indices = [
                    *TopologicalSorter(self._reverse_monotonicity_graph).static_order()
                ]
            except CycleError as exc:
                raise ValueError("monotonicity_pairs is cyclic") from exc
        self.kernel_init = kernel_init

        self.kernel = torch.nn.Parameter(torch.Tensor(self.num_categories, 1).double())
        if kernel_init == CategoricalCalibratorInit.CONSTANT:
            if output_min is not None and output_max is not None:
                init_value = (output_min + output_max) / 2
            elif output_min is not None:
                init_value = output_min
            elif output_max is not None:
                init_value = output_max
            else:
                init_value = 0.0
            torch.nn.init.constant_(self.kernel, init_value)
        elif kernel_init == CategoricalCalibratorInit.UNIFORM:
            if output_min is None and output_max is None:
                low, high = -0.05, 0.05
            elif output_min is None:
                low, high = output_max - 0.05, output_max
            elif output_max is None:
                low, high = output_min, output_min + 0.05
            else:
                low, high = output_min, output_max
            torch.nn.init.uniform_(self.kernel, low, high)
        else:
            raise ValueError(f"Unknown kernel init: {kernel_init}")

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=invalid-name
        """Calibrates categorical inputs through a learned mapping.

        Args:
            x: The input tensor of category indices of shape `(batch_size, 1)`.

        Returns:
            torch.Tensor of shape `(batch_size, 1)` containing calibrated input values.
        """
        if self.missing_input_value is not None:
            missing_category_tensor = torch.zeros_like(x) + (self.num_categories - 1)
            x = torch.where(x == self.missing_input_value, missing_category_tensor, x)
        # TODO: test if using torch.gather is faster than one-hot matmul.
        one_hot = torch.nn.functional.one_hot(
            torch.squeeze(x, -1).long(), num_classes=self.num_categories
        ).double()
        return torch.mm(one_hot, self.kernel)

    @torch.no_grad()
    def assert_constraints(self, eps=1e-6) -> List[str]:
        """Asserts that layer satisfies specified constraints.

        This checks that weights at the indexes of monotonicity pairs are in the correct
        order and that the output is within bounds.

        Args:
            eps: the margin of error allowed

        Returns:
            A list of messages describing violated constraints including violated
            monotonicity pairs. If no constraints  violated, the list will be empty.
        """
        weights = torch.squeeze(self.kernel.data)
        messages = []

        if self.output_max is not None and torch.max(weights) > self.output_max + eps:
            messages.append("Max weight greater than output_max.")
        if self.output_min is not None and torch.min(weights) < self.output_min - eps:
            messages.append("Min weight less than output_min.")

        if self.monotonicity_pairs:
            violation_indices = [
                (i, j)
                for (i, j) in self.monotonicity_pairs
                if weights[i] - weights[j] > eps
            ]
            if violation_indices:
                messages.append(f"Monotonicity violated at: {str(violation_indices)}.")

        return messages

    @torch.no_grad()
    def constrain(self) -> None:
        """Projects kernel into desired constraints."""
        projected_kernel_data = self.kernel.data
        if self.monotonicity_pairs:
            projected_kernel_data = self._approximately_project_monotonicity_pairs(
                projected_kernel_data
            )
        if self.output_min is not None:
            projected_kernel_data = torch.maximum(
                projected_kernel_data, torch.tensor(self.output_min)
            )
        if self.output_max is not None:
            projected_kernel_data = torch.minimum(
                projected_kernel_data, torch.tensor(self.output_max)
            )
        self.kernel.data = projected_kernel_data

    @torch.no_grad()
    def keypoints_inputs(self) -> torch.Tensor:
        """Returns a tensor of keypoint inputs (category indices)."""
        if self.missing_input_value is not None:
            return torch.cat(
                (
                    torch.arange(self.num_categories - 1),
                    torch.tensor([self.missing_input_value]),
                ),
                0,
            )
        return torch.arange(self.num_categories)

    @torch.no_grad()
    def keypoints_outputs(self) -> torch.Tensor:
        """Returns a tensor of keypoint outputs."""
        return torch.squeeze(self.kernel.data, -1)

    ################################################################################
    ############################## PRIVATE METHODS #################################
    ################################################################################

    def _approximately_project_monotonicity_pairs(self, kernel_data) -> torch.Tensor:
        """Projects kernel such that the monotonicity pairs are satisfied.

        The kernel will be projected such that `kernel_data[i] <= kernel_data[j]`. This
        results in calibrated outputs that adhere to the desired constraints.

        Args:
            kernel_data: The tensor of shape `(self.num_categories, 1)` to be projected
                into the constraints specified by `self.monotonicity pairs`.

        Returns:
            Projected kernel data. To prevent the kernel from drifting in one direction,
            the data returned is the average of the min/max and max/min projections.
        """
        projected_kernel_data = torch.unbind(kernel_data, 0)

        def project(data, monotonicity_graph, step, minimum):
            projected_data = list(data)
            sorted_indices = self._monotonically_sorted_indices
            if minimum:
                sorted_indices = sorted_indices[::-1]
            for i in sorted_indices:
                if i in monotonicity_graph:
                    projection = projected_data[i]
                    for j in monotonicity_graph[i]:
                        if minimum:
                            projection = torch.minimum(projection, projected_data[j])
                        else:
                            projection = torch.maximum(projection, projected_data[j])
                        if step == 1.0:
                            projected_data[i] = projection
                        else:
                            projected_data[i] = (
                                step * projection + (1 - step) * projected_data[i]
                            )
            return projected_data

        projected_kernel_min_max = project(
            projected_kernel_data, self._monotonicity_graph, 0.5, minimum=True
        )
        projected_kernel_min_max = project(
            projected_kernel_min_max,
            self._reverse_monotonicity_graph,
            1.0,
            minimum=False,
        )
        projected_kernel_min_max = torch.stack(projected_kernel_min_max)

        projected_kernel_max_min = project(
            projected_kernel_data, self._reverse_monotonicity_graph, 0.5, minimum=False
        )
        projected_kernel_max_min = project(
            projected_kernel_max_min, self._monotonicity_graph, 1.0, minimum=True
        )
        projected_kernel_max_min = torch.stack(projected_kernel_max_min)

        return (projected_kernel_min_max + projected_kernel_max_min) / 2



"""PyTorch Calibrated Models to easily implement common calibrated model architectures.

PyTorch Calibrated Models make it easy to construct common calibrated model
architectures. To construct a PyTorch Calibrated Model, pass a calibrated modeling
config to the corresponding calibrated model.
"""
from typing import Dict, List, Optional, Union

import numpy as np
import torch

from .enums import (
    CategoricalCalibratorInit,
    FeatureType,
    Monotonicity,
    NumericalCalibratorInit,
)
from .features import CategoricalFeature, NumericalFeature
from .layers.categorical_calibrator import CategoricalCalibrator
from .layers.linear import Linear
from .layers.numerical_calibrator import NumericalCalibrator


# pylint: disable-next=too-many-instance-attributes
class CalibratedLinear(torch.nn.Module):
    """PyTorch Calibrated Linear Model.

    Creates a `torch.nn.Module` representing a calibrated linear model, which will be
    constructed using the provided model configuration. Note that the model inputs
    should match the order in which they are defined in the `feature_configs`.

    Attributes:
        - All `__init__` arguments.
        calibrators: A dictionary that maps feature names to their calibrators.
        linear: The `Linear` layer of the model.
        output_calibrator: The output `NumericalCalibrator` calibration layer. This
            will be `None` if no output calibration is desired.

    Example:

    ```python
    csv_data = CSVData(...)

    feature_configs = [...]
    calibrated_model = CalibratedLinear(feature_configs, ...)

    loss_fn = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(calibrated_model.parameters(recurse=True), lr=1e-1)

    csv_data.prepare(feature_configs, "target", ...)
    for epoch in range(100):
        for examples, targets in csv_data.batch(64):
            optimizer.zero_grad()
            outputs = calibrated_model(inputs)
            loss = loss_fn(outputs, labels)
            loss.backward()
            optimizer.step()
            calibrated_model.constrain()
    ```
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        features: List[Union[NumericalFeature, CategoricalFeature]],
        output_min: Optional[float] = None,
        output_max: Optional[float] = None,
        use_bias: bool = True,
        output_calibration_num_keypoints: Optional[int] = None,
    ) -> None:
        """Initializes an instance of `CalibratedLinear`.

        Args:
            features: A list of numerical and/or categorical feature configs.
            output_min: The minimum output value for the model. If `None`, the minimum
                output value will be unbounded.
            output_max: The maximum output value for the model. If `None`, the maximum
                output value will be unbounded.
            use_bias: Whether to use a bias term for the linear combination. If any of
                `output_min`, `output_max`, or `output_calibration_num_keypoints` are
                set, a bias term will not be used regardless of the setting here.
            output_calibration_num_keypoints: The number of keypoints to use for the
                output calibrator. If `None`, no output calibration will be used.

        Raises:
            ValueError: If any feature configs are not `NUMERICAL` or `CATEGORICAL`.
        """
        super().__init__()

        self.features = features
        self.output_min = output_min
        self.output_max = output_max
        self.use_bias = use_bias
        self.output_calibration_num_keypoints = output_calibration_num_keypoints

        linear_monotonicities = []
        self.calibrators = torch.nn.ModuleDict()
        for feature in features:
            if feature.feature_type == FeatureType.NUMERICAL:
                self.calibrators[feature.feature_name] = NumericalCalibrator(
                    input_keypoints=feature.input_keypoints,
                    missing_input_value=feature.missing_input_value,
                    output_min=output_min,
                    output_max=output_max,
                    monotonicity=feature.monotonicity,
                    kernel_init=NumericalCalibratorInit.EQUAL_SLOPES,
                    projection_iterations=feature.projection_iterations,
                )
                if feature.monotonicity == Monotonicity.NONE:
                    linear_monotonicities.append(Monotonicity.NONE)
                else:
                    linear_monotonicities.append(Monotonicity.INCREASING)
            elif feature.feature_type == FeatureType.CATEGORICAL:
                self.calibrators[feature.feature_name] = CategoricalCalibrator(
                    num_categories=len(feature.categories),
                    missing_input_value=feature.missing_input_value,
                    output_min=output_min,
                    output_max=output_max,
                    monotonicity_pairs=feature.monotonicity_index_pairs,
                    kernel_init=CategoricalCalibratorInit.UNIFORM,
                )
                if not feature.monotonicity_pairs:
                    linear_monotonicities.append(Monotonicity.NONE)
                else:
                    linear_monotonicities.append(Monotonicity.INCREASING)
            else:
                raise ValueError(
                    f"Unknown feature type {feature.feature_type} for feature "
                    f"{feature.feature_name}"
                )

        self.linear = Linear(
            input_dim=len(features),
            monotonicities=linear_monotonicities,
            use_bias=use_bias,
            weighted_average=(
                output_min is not None
                or output_max is not None
                or output_calibration_num_keypoints
            ),
        )

        self.output_calibrator = None
        if output_calibration_num_keypoints:
            non_monotonic = all(m == Monotonicity.NONE for m in linear_monotonicities)
            self.output_calibrator = NumericalCalibrator(
                input_keypoints=np.linspace(
                    0.0, 1.0, num=output_calibration_num_keypoints
                ),
                missing_input_value=None,
                output_min=output_min,
                output_max=output_max,
                monotonicity=Monotonicity.NONE
                if non_monotonic
                else Monotonicity.INCREASING,
                kernel_init=NumericalCalibratorInit.EQUAL_HEIGHTS,
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=invalid-name
        """Runs an input through the network to produce a calibrated linear output.

        Args:
            x: The input tensor of feature values of shape `(batch_size, num_features)`.

        Returns:
            torch.Tensor of shape `(batch_size, 1)` containing the model output result.
        """
        result = torch.column_stack(
            tuple(
                calibrator(x[:, i, None])
                for i, calibrator in enumerate(self.calibrators.values())
            )
        )
        result = self.linear(result)
        if self.output_calibrator is not None:
            result = self.output_calibrator(result)

        return result

    @torch.no_grad()
    def assert_constraints(self) -> Dict[str, List[str]]:
        """Asserts all layers within model satisfied specified constraints.

        Asserts monotonicity pairs and output bounds for categorical calibrators,
        monotonicity and output bounds for numerical calibrators, and monotonicity and
        weights summing to 1 if weighted_average for linear layer.

        Returns:
            A dict where key is feature_name for calibrators and 'linear' for the linear
            layer, and value is the error messages for each layer. Layers with no error
            messages are not present in the dictionary.
        """
        messages = {}

        for name, calibrator in self.calibrators.items():
            calibrator_messages = calibrator.assert_constraints()
            if calibrator_messages:
                messages[name] = calibrator_messages
        linear_messages = self.linear.assert_constraints()
        if linear_messages:
            messages["linear"] = linear_messages

        return messages

    @torch.no_grad()
    def constrain(self) -> None:
        """Constrains the model into desired constraints specified by the config."""
        for calibrator in self.calibrators.values():
            calibrator.constrain()
        self.linear.constrain()
        if self.output_calibrator:
            self.output_calibrator.constrain()



"""A Pipeline for calibrated modeling."""
from __future__ import annotations

import logging
import os
import pickle
from time import sleep
from typing import Dict, List, Optional, Tuple, Union
import itertools
import numpy as np
import pandas as pd

from .api import (
    download_prepared_dataset,
    get_api_key,
    get_dataset_uuids,
    get_inference_results,
    get_inference_status,
    get_trained_model_uuids,
    post_inference,
    post_pipeline,
    post_pipeline_config,
    post_pipeline_feature_configs,
    post_trained_model,
    post_trained_model_analysis,
    post_dataset,
    get_pipeline,
    post_hypertune_job,
)
from .constants import INFERENCE_POLLING_INTERVAL, SOTAI_BASE_URL
from .data import determine_feature_types, replace_missing_values
from .enums import (
    APIStatus,
    FeatureType,
    InferenceConfigStatus,
    LossType,
    Metric,
    TargetType,
)
from .trained_model import TrainedModel
from .training import train_and_evaluate_model
from .types import (
    CategoricalFeatureConfig,
    Dataset,
    DatasetSplit,
    LinearConfig,
    NumericalFeatureConfig,
    PipelineConfig,
    PreparedData,
    TrainingConfig,
    HypertuneConfig,
)


class Pipeline:  # pylint: disable=too-many-instance-attributes
    """A pipeline for calibrated modeling.

    The pipeline defines the configuration for training a calibrated model. The
    pipeline itself defines the features, target, and target type to be used. When
    training a model, the data and configuration used will be versioned and stored in
    the pipeline. The pipeline can be used to train multiple models with different
    configurations if desired; however, the target, target type, and primary metric
    should not be changed after initialization so that models trained by this pipeline
    can be compared.

    Example:
    ```python
    data = pd.read_csv(...)
    pipeline = Pipeline(features, target, TargetType.CLASSIFICATION)
    trained_model = pipeline.train(data)
    ```

    Attributes:
        name: The name of the pipeline.
        target: The name of the target column.
        target_type: The type of the target column.
        primary_metric: The primary metric to use for training and evaluation.
        feature_configs: A dictionary mapping feature names to feature configurations.
        shuffle_data: Whether or not to shuffle the data before training.
        drop_empty_percentage: The percentage of empty values in a single row that will
            result in dropping the row before training.
        dataset_split: The dataset split to use for training and evaluation.
        configs: A dictionary mapping versioned pipeline configuration IDs to pipeline
            configurations.
        datasets: A dictionary mapping versioned dataset IDs to datasets.
        uuid: The UUID of the pipeline. This will be `None` unless the pipeline has been
            published.
        trained_models: A dictionary mapping trained model UUIDs to trained models.
    """

    def __init__(
        self,
        features: List[str],
        target: str,
        target_type: TargetType,
        categories: Optional[Dict[str, Union[List[int], List[str]]]] = None,
        primary_metric: Optional[Metric] = None,
        name: Optional[str] = None,
        default_allow_hosting: bool = True,
    ):
        """Initializes an instance of `Pipeline`.

        The pipeline is initialized with a default config, which can be modified later.
        The target type can be optionally specfified. The default primary metric will be
        AUC for classification and Mean Squared Error for regression if not specified.

        Args:
            features: The column names in your data to use as features.
            target: The name of the target column.
            target_type: The type of the target column.
            categories: A dictionary mapping feature names to unique categories. Any
                values not in the categories list for a given feature will be treated
                as a missing value.
            primary_metric: The primary metric to use for training and evaluation.
            name: The name of the pipeline. If not provided, the name will be set to
                `{target}_{target_type}`.
            default_allow_hosting: Whether or not to allow hosting by default for
                datasets and models created by this pipeline.
        """
        self.name: str = name if name else f"{target}_{target_type}"
        self.target: str = target
        self.target_type: TargetType = target_type
        self.default_allow_hosting: bool = default_allow_hosting

        self.primary_metric: Metric = (
            primary_metric
            if primary_metric is not None
            else (
                Metric.AUC
                if self.target_type == TargetType.CLASSIFICATION
                else Metric.MSE
            )
        )
        self.feature_configs: Dict[
            str, Union[CategoricalFeatureConfig, NumericalFeatureConfig]
        ] = {
            feature_name: (
                CategoricalFeatureConfig(
                    name=feature_name,
                    categories=categories[feature_name],
                )
                if categories and feature_name in categories
                else NumericalFeatureConfig(name=feature_name)
            )
            for feature_name in features
        }
        self.shuffle_data: bool = True
        self.drop_empty_percentage: int = 70
        self.dataset_split: DatasetSplit = DatasetSplit(train=80, val=10, test=10)

        # Maps a pipeline config id to its corresponding `PipelineConfig`` instance.
        self.configs: Dict[int, PipelineConfig] = {}
        # Maps a dataset id to its corresponding `Dataset`` instance.
        self.datasets: Dict[int, Dataset] = {}

        # Maps a trained model id to its corresponding `TrainedModel`` instance.
        self.trained_models: Dict[int, TrainedModel] = {}

        # Tracking for the next versioned config, dataset, and model.
        self._next_config_id = 0
        self._next_dataset_id = 0
        self._next_model_id = 0

        # Tracks
        self.uuid = None

    def prepare(  # pylint: disable=too-many-locals
        self,
        data: pd.DataFrame,
        pipeline_config_id: Optional[int] = None,
    ) -> Tuple[Dataset, PipelineConfig]:
        """Prepares the data and versions it along with the current pipeline config.

        If any features in data are detected as non-numeric, the pipeline will attempt
        to handle them as categorical features. Any features that the pipeline cannot
        handle will be skipped.

        Args:
            data: The raw data to be prepared for training.
            pipeline_config_id: The id of the pipeline config to be used for training.
                If not provided, the current pipeline config will be used and versioned.

        Returns:
            A tuple of the versioned dataset and pipeline config.
        """
        data.replace("", np.nan, inplace=True)  # treat empty strings as NaN
        if pipeline_config_id is None:
            pipeline_config = self._version_pipeline_config(data)
        else:
            pipeline_config = self.configs[pipeline_config_id]

        if (
            pipeline_config.target_type == TargetType.CLASSIFICATION
            and data[pipeline_config.target].dtype != np.int64
            and data[pipeline_config.target].unique().sort() != [0, 1]
        ):
            raise ValueError(
                "Target column must be of type int64 with values [0,1] for classification."
            )

        # Select only the features defined in the pipeline config.
        data = data[list(pipeline_config.feature_configs.keys()) + [self.target]]
        # Drop rows with too many missing values according to the drop empty percent.
        max_num_empty_columns = int(
            (pipeline_config.drop_empty_percentage * data.shape[1]) / 100
        )
        data = data[data.isnull().sum(axis=1) <= max_num_empty_columns]
        # Replace any missing values (i.e. NaN) with missing value constants.
        data = replace_missing_values(data, pipeline_config.feature_configs)
        if pipeline_config.shuffle_data:
            data = data.sample(frac=1).reset_index(drop=True)
        train_percentage = pipeline_config.dataset_split.train / 100
        train_data = data.iloc[: int(len(data) * train_percentage)]
        val_percentage = pipeline_config.dataset_split.val / 100
        val_data = data.iloc[
            int(len(data) * train_percentage) : int(
                len(data) * (train_percentage + val_percentage)
            )
        ]
        test_data = data.iloc[int(len(data) * (train_percentage + val_percentage)) :]

        dataset_id = self._next_dataset_id
        self._next_dataset_id += 1
        dataset = Dataset(
            id=dataset_id,
            pipeline_config_id=pipeline_config.id,
            prepared_data=PreparedData(train=train_data, val=val_data, test=test_data),
            allow_hosting=self.default_allow_hosting,
        )
        self.datasets[dataset_id] = dataset

        return dataset, pipeline_config

    def train(
        self,
        data: Union[pd.DataFrame, int],
        pipeline_config_id: Optional[int] = None,
        model_config: Optional[LinearConfig] = None,
        training_config: Optional[TrainingConfig] = None,
    ) -> TrainedModel:
        """Returns a calibrated model trained according to the given configs.

        Args:
            data: The raw data to be prepared and trained on. If an int is provided,
                it is assumed to be a dataset id and the corresponding dataset will be
                used.
            pipeline_config_id: The id of the pipeline config to be used for training.
                If not provided, the current pipeline config will be versioned and used.
                If data is an int, this argument is ignored and the pipeline config used
                to prepare the data with the given id will be used.
            model_config: The config to be used for training the model. If not provided,
                a default config will be used.
            training_config: The config to be used for training the model. If not
                provided, a default config will be used.

        Returns:
            A `TrainedModel` instance.
        """
        if isinstance(data, int):
            dataset = self.datasets[data]
            pipeline_config = self.configs[dataset.pipeline_config_id]
        else:
            dataset, pipeline_config = self.prepare(data, pipeline_config_id)

        if model_config is None:
            model_config = LinearConfig()

        if training_config is None:
            training_config = TrainingConfig(
                loss_type=LossType.BINARY_CROSSENTROPY
                if self.target_type == TargetType.CLASSIFICATION
                else LossType.MSE
            )

        model, training_results = train_and_evaluate_model(
            dataset,
            self.target,
            self.primary_metric,
            pipeline_config,
            model_config,
            training_config,
        )
        trained_model = TrainedModel(
            id=self._next_model_id,
            dataset_id=dataset.id,
            pipeline_config=pipeline_config,
            model_config=model_config,
            training_config=training_config,
            training_results=training_results,
            model=model,
            allow_hosting=self.default_allow_hosting,
        )

        self.trained_models[self._next_model_id] = trained_model
        self._next_model_id += 1
        return trained_model

    def hypertune(
        self,
        data: Union[int, pd.DataFrame],
        hypertune_config: HypertuneConfig,
        pipeline_config_id: Optional[int] = None,
        model_config: Optional[LinearConfig] = None,
        hosted: bool = False,
    ) -> List[Union[TrainedModel, str]]:
        """Returns a list of trained models trained according to the given configs.

        Args:
            data: The raw data to be prepared and trained on. If an int is provided,
                it is assumed to be a dataset id and the corresponding dataset will be
                used.
            hypertune_config: The config to be used for hypertuning the model.
            pipeline_config_id: The id of the pipeline config to be used for training.
                If not provided, the current pipeline config will be versioned and used.
                If data is an int, this argument is ignored and the pipeline config used
                to prepare the data with the given id will be used.
            model_config: The config to be used for training the model. If not provided,
                a default config will be used.
            hosted: Whether to run the hypertune job on the SOTAI cloud. If False, the
                hypertune job will be run locally.

        Returns:
            A list of `TrainedModel` instances if run locally, or a list of trained
            model uuids if run in the SOTAI cloud.
        """

        if pipeline_config_id is None:
            pipeline_config = self._version_pipeline_config(data)
            pipeline_config_id = pipeline_config.id
        else:
            pipeline_config = self.configs[pipeline_config_id]

        if model_config is None:
            model_config = LinearConfig()

        if hosted:
            if not get_api_key():
                raise ValueError(
                    "You must have an API key to run hosted hypertuning."
                    " Please visit app.sotai.ai to get an API key."
                )

            _ = self._post_pipeline_config(pipeline_config=pipeline_config)

            dataset_uuid = self._upload_dataset(self.uuid, data, pipeline_config_id)
            if dataset_uuid is None:
                return []

            hypertune_response, trained_model_uuids = post_hypertune_job(
                hypertune_config,
                pipeline_config,
                model_config,
                dataset_uuid,
                self._next_model_id,
            )
            if hypertune_response == APIStatus.ERROR:
                return []

            self._next_model_id += len(trained_model_uuids)
            return trained_model_uuids

        return self._local_hypertuning(
            hypertune_config=hypertune_config,
            data=data,
            pipeline_config_id=pipeline_config.id,
            model_config=model_config,
        )

    def publish(self) -> Optional[str]:
        """Uploads the pipeline to the SOTAI web client.

        Returns:
            If the pipeline was successfully uploaded, the pipeline UUID.
            Otherwise, None.
        """
        pipeline_response_status, pipeline_uuid = post_pipeline(self)
        if pipeline_response_status == APIStatus.ERROR:
            return None

        self.uuid = pipeline_uuid
        return self.uuid

    def analysis(  # pylint: disable=too-many-return-statements
        self,
        trained_model: TrainedModel,
    ) -> Optional[str]:
        """Charts the results for the specified trained model in the SOTAI web client.

        This function requires an internet connection and a SOTAI account. The trained
        model will be uploaded to the SOTAI web client for analysis.

        If you would like to analyze the results for a trained model without uploading
        it to the SOTAI web client, the data is available in `training_results`.

        Args:
            trained_model: The trained model to be analyzed.

        Returns:
            If the analysis was successfully run, the analysis URL. Otherwise `None`.
        """
        if trained_model.analysis_url:  # early exit if analysis has already been run.
            return trained_model.analysis_url

        if not get_api_key():
            raise ValueError(
                "You must have an API key to run analysis."
                " Please visit app.sotai.ai to get an API key."
            )

        pipeline_config_uuid = self._post_pipeline_config(trained_model.pipeline_config)

        if not pipeline_config_uuid:
            return None

        analysis_response_status, analysis_results = post_trained_model_analysis(
            pipeline_config_uuid, trained_model
        )

        if analysis_response_status == APIStatus.ERROR:
            return None

        trained_model.uuid = analysis_results["trained_model_metadata_uuid"]

        upload_response = self._upload_model(trained_model)
        if upload_response == APIStatus.ERROR:
            return None

        # TODO: update to use response analysisUrl once no longer broken.
        analysis_url = (
            f"{SOTAI_BASE_URL}/pipelines/{self.uuid}"
            f"/trained-models/{trained_model.uuid}/overall-model-results"
        )
        trained_model.analysis_url = analysis_url

        return analysis_url

    def inference(
        self,
        filepath: str,
        trained_model: TrainedModel,
    ) -> Optional[str]:
        """Runs inference on a dataset with a trained model in the SOTAI cloud.

        Args:
            inference_dataset_path: The path to the dataset to run inference on.
            trained_model: The trained model to use for inference.

        Returns:
            If UUID of the inference run. If unsuccessful, `None`.
        """

        if not get_api_key():
            raise ValueError(
                "You must have an API key to run inference."
                " Please visit app.sotai.ai to get an API key."
            )

        if trained_model.uuid is None:
            self.analysis(trained_model)

        data = pd.read_csv(filepath)
        expected_columns = list(trained_model.pipeline_config.feature_configs.keys())
        if not set(expected_columns) <= set(data.columns):
            raise ValueError(
                "Inference data must have columns for all features to run inference."
            )

        inference_response_status, inference_uuid = post_inference(
            filepath, trained_model.uuid
        )
        if inference_response_status == APIStatus.ERROR:
            return None

        return inference_uuid

    def await_inference(
        self,
        inference_uuid: str,
        inference_results_folder_path: str,
    ):
        """Polls the SOTAI cloud for the results of the specified inference job.

        Args:
            inference_uuid: The uuid of the inference job to poll.
            inference_results_folder_path: The path to save the inference results to.

        Returns:
            If inference was successfully run, the path to the inference results.
        """
        while True:
            inference_response_status, inference_job_status = get_inference_status(
                inference_uuid
            )
            logging.info("Current inference job status: %s", inference_job_status)
            if (
                inference_response_status == APIStatus.SUCCESS
                and inference_job_status == InferenceConfigStatus.SUCCESS
            ):
                get_inference_response = get_inference_results(
                    inference_uuid, inference_results_folder_path
                )
                if get_inference_response == APIStatus.ERROR:
                    logging.info("Error getting inference results")
                else:
                    logging.info(
                        "Inference results saved to: %s ", inference_results_folder_path
                    )
                return inference_results_folder_path

            sleep(INFERENCE_POLLING_INTERVAL)

    def save(self, filepath: str, include_trained_models: bool = True):
        """Saves the pipeline to the specified filepath.

        Trained models will be saved with the pipeline by default.

        Args:
            filepath: The directory to which the pipeline wil be saved. If the directory
                does not exist, this function will attempt to create it. If the
                directory already exists, this function will overwrite any existing
                content with conflicting filenames.
            include_trained_models: If True (default), then all models trained under
                this pipeline present in the trained_models dictionary attribute will
                be stored along with the pipeline under a `trained_models/{id}`
                directory. If False, trained models will be excluded from saving.
        """
        if not os.path.exists(filepath):
            os.makedirs(filepath)
        with open(os.path.join(filepath, "pipeline.pkl"), "wb") as file:
            pickle.dump(self, file)

        if include_trained_models:
            for trained_model in self.trained_models.values():
                # trained_model.save(filepath) handles creating the directory.
                trained_model.save(
                    os.path.join(filepath, f"trained_models/{trained_model.id}")
                )

    def sync(self):
        """Syncs the pipeline with the SOTAI cloud."""
        if self.uuid is None:
            self.publish()

        for pipeline_config in self.configs.values():
            if pipeline_config.uuid is None and pipeline_config.allow_hosting:
                self._post_pipeline_config(pipeline_config)

        for trained_model in self.trained_models.values():
            if trained_model.uuid is None and trained_model.allow_hosting:
                self.analysis(trained_model)
                self._upload_model(trained_model)

        for dataset in self.datasets.values():
            if dataset.uuid is None and dataset.allow_hosting:
                self._upload_dataset(self.uuid, dataset.id, dataset.pipeline_config_id)

        trained_model_uuids = get_trained_model_uuids(self.uuid)
        local_trained_model_uuids = [
            trained_model.uuid for trained_model in self.trained_models.values()
        ]
        for trained_model_uuid in trained_model_uuids:
            if trained_model_uuid not in local_trained_model_uuids:
                trained_model = TrainedModel.from_hosted(trained_model_uuid)
                if trained_model is not None:
                    self.trained_models[trained_model.id] = trained_model

    @classmethod
    def from_hosted(
        cls,
        pipeline_uuid: str,
        include_trained_models: bool = False,
        include_datasets: bool = False,
    ) -> Pipeline:
        """Returns a new pipeline created from the specified hosted pipeline uuid.

        Args:
            pipeline_uuid: The uuid of the hosted pipeline to load.
            include_trained_models: If True, trained models will be loaded along with
                the pipeline. If False (default), trained models will not be loaded.

        Returns:
            A `Pipeline` instance.
        """
        (
            pipeline_metadata,
            last_config_id,
            pipeline_configs,
            trained_model_uuids,
        ) = get_pipeline(pipeline_uuid)

        if not pipeline_configs:
            raise ValueError(
                "Pipeline configs not found. Please run training before loading."
            )
        pipeline = Pipeline.from_config(
            name=pipeline_metadata["name"],
            config=pipeline_configs[last_config_id],
        )
        pipeline.configs = pipeline_configs
        pipeline.uuid = pipeline_uuid

        if include_trained_models:
            for trained_model_uuid in trained_model_uuids:
                trained_model = TrainedModel.from_hosted(trained_model_uuid)
                if trained_model:
                    pipeline.trained_models[trained_model.id] = trained_model

        if include_datasets:
            _, dataset_uuids = get_dataset_uuids(pipeline_uuid)
            for dataset_uuid in dataset_uuids:
                _, dataset = download_prepared_dataset(dataset_uuid)
                if dataset is not None:
                    pipeline.datasets[dataset.id] = dataset

        pipeline._next_config_id = len(pipeline.configs) + 1
        if len(pipeline.trained_models) != 0:
            pipeline._next_model_id = max(pipeline.trained_models.keys()) + 1

        return pipeline

    @classmethod
    def load(cls, filepath: str) -> Pipeline:
        """Loads the pipeline from the specified filepath.

        Args:
            filepath: The filepath from which to load the pipeline. The filepath should
                point to a file created by the `save` method of a `TrainedModel`
                instance.

        Returns:
            A `Pipeline` instance.
        """
        with open(os.path.join(filepath, "pipeline.pkl"), "rb") as file:
            pipeline = pickle.load(file)

        trained_model_dir = os.path.join(filepath, "trained_models")
        if os.path.isdir(trained_model_dir):
            for directory in os.listdir(trained_model_dir):
                if os.path.isdir(os.path.join(trained_model_dir, directory)):
                    trained_model = TrainedModel.load(
                        os.path.join(trained_model_dir, directory)
                    )
                    pipeline.trained_models[trained_model.id] = trained_model

        return pipeline

    @classmethod
    def from_config(
        cls, config: PipelineConfig, name: Optional[str] = None
    ) -> Pipeline:
        """Returns a new pipeline created from the specified config."""
        pipeline = cls(
            features=list(config.feature_configs.keys()),
            target=config.target,
            target_type=config.target_type,
            primary_metric=config.primary_metric,
            name=name,
        )
        pipeline.feature_configs = config.feature_configs
        pipeline.shuffle_data = config.shuffle_data
        pipeline.drop_empty_percentage = config.drop_empty_percentage
        pipeline.dataset_split = config.dataset_split

        return pipeline

    ############################################################################
    #                            Private Methods                               #
    ############################################################################

    def __getstate__(self):
        """Return state values to be pickled.

        We exclude the trained models from the pipeline.
        """
        state = self.__dict__.copy()
        del state["trained_models"]
        return state

    def __setstate__(self, state):
        """Restore state from the unpickled state values.

        We restore the trained model field
        """
        self.__dict__.update(state)
        self.trained_models = {}

    def _version_pipeline_config(self, data: pd.DataFrame):
        """Returns a new `PipelineConfig` instance verisoned from the current config."""
        pipeline_config_id = self._next_config_id
        self._next_config_id += 1
        pipeline_config = PipelineConfig(
            id=pipeline_config_id,
            target=self.target,
            target_type=self.target_type,
            primary_metric=self.primary_metric,
            feature_configs=self.feature_configs,
            shuffle_data=self.shuffle_data,
            drop_empty_percentage=self.drop_empty_percentage,
            dataset_split=self.dataset_split,
            allow_hosting=self.default_allow_hosting,
        )

        feature_types = determine_feature_types(
            data[list(pipeline_config.feature_configs.keys())]
        )
        for feature_name, feature_type in feature_types.items():
            feature_config = pipeline_config.feature_configs[feature_name]
            if (
                feature_type == FeatureType.NUMERICAL
                or feature_config.type == FeatureType.CATEGORICAL
            ):
                continue
            if feature_type == FeatureType.CATEGORICAL:
                logging.info(
                    "Detected %s as categorical. Replacing numerical config with "
                    "default categorical config using unique values as categories",
                    feature_name,
                )
                pipeline_config.feature_configs[
                    feature_name
                ] = CategoricalFeatureConfig(
                    name=feature_name,
                    categories=sorted(data[feature_name].dropna().unique().tolist()),
                )
            else:
                logging.info(
                    "Removing feature %s because its data type is not supported.",
                    feature_name,
                )
                pipeline_config.feature_configs.pop(feature_name)

        self.configs[pipeline_config_id] = pipeline_config
        return pipeline_config

    def _upload_model(
        self,
        trained_model: TrainedModel,
    ) -> APIStatus:
        """Uploads the trained model to the SOTAI web client.

        If a model has already been uploaded, this function will return without
        uploading the model again. This function requires an internet connection and a
        SOTAI account. The trained model will be uploaded to the SOTAI web client for
        inference.

        Args:
            trained_model: The trained model to upload.
        """

        if not get_api_key():
            raise ValueError(
                "You must have an API key to upload a model."
                " Please visit app.sotai.ai to get an API key."
            )
        model_save_folder_path = f"/tmp/sotai/model/{trained_model.uuid}"

        if trained_model.uuid is None:
            raise ValueError("Must run analysis to generate uuid before uploading.")

        trained_model.save(model_save_folder_path)
        trained_model_response = post_trained_model(
            model_save_folder_path, trained_model.uuid
        )
        return trained_model_response

    def _upload_dataset(
        self,
        pipeline_uuid: str,
        data: Union[int, pd.DataFrame],
        pipeline_config_id: int,
    ) -> Optional[str]:
        """Uploads the dataset to the SOTAI web client.

        If a dataset has already been uploaded, this function will return without
        uploading the dataset again. This function requires an internet connection and a
        SOTAI account. The dataset will be uploaded to the SOTAI web client for
        inference.

        Args:
            pipeline_uuid: The uuid of the pipeline to upload the dataset to.
            data: The raw data to be prepared and trained on. If an int is provided,
                it is assumed to be a dataset id and the corresponding dataset will be
                used.
            pipeline_config_id: The id of the pipeline config to be used for training.
                If not provided, the current pipeline config will be versioned and used.
                If data is an int, this argument is ignored and the pipeline config used
                to prepare the data with the given id will be used.

        Returns:
            If the dataset was successfully uploaded, the dataset UUID.
            Otherwise, `None`.
        """
        filepath = "/tmp/sotai"
        if not os.path.exists(filepath):
            os.makedirs("/tmp/sotai/")

        if isinstance(data, int):
            dataset = self.datasets[data]
            if dataset.uuid is not None:
                return dataset.uuid
        else:
            dataset, new_pipeline_config = self.prepare(data, pipeline_config_id)
            pipeline_config_id = new_pipeline_config.id

        if self.configs[pipeline_config_id].uuid is None:
            status, pipeline_config_uuid = post_pipeline_config(
                pipeline_uuid, self.configs[pipeline_config_id]
            )
            if status == APIStatus.ERROR:
                return None
        else:
            pipeline_config_uuid = self.configs[pipeline_config_id].uuid

        dataset.prepared_data.train.to_csv("/tmp/sotai/train.csv")
        dataset.prepared_data.test.to_csv("/tmp/sotai/test.csv")
        dataset.prepared_data.val.to_csv("/tmp/sotai/validation.csv")
        columns = dataset.prepared_data.train.columns.tolist()
        _, dataset_uuid = post_dataset(
            "/tmp/sotai/train.csv",
            "/tmp/sotai/test.csv",
            "/tmp/sotai/validation.csv",
            columns=columns,
            categorical_columns=[],
            pipeline_config_uuid=pipeline_config_uuid,
            dataset_id=dataset.id,
        )
        dataset.uuid = dataset_uuid

        return dataset_uuid

    def _post_pipeline_config(self, pipeline_config: PipelineConfig) -> str:
        """Posts a pipeline config to the SOTAI web client.

        If a pipeline config has already been posted, this function will return without
        posting the config again.

        Args:
            pipeline_config: The pipeline config to post.

        Returns:
            If the pipeline config was successfully posted, the pipeline config UUID.
            Otherwise, None.
        """

        if pipeline_config.uuid is not None:
            return pipeline_config.uuid

        if self.uuid is None:
            self.uuid = self.publish()

        if self.uuid is None:
            return None

        pipeline_config_response_status, pipeline_config_uuid = post_pipeline_config(
            self.uuid, pipeline_config
        )

        if pipeline_config_response_status == APIStatus.ERROR:
            return None

        pipeline_config.uuid = pipeline_config_uuid

        feature_config_response_status = post_pipeline_feature_configs(
            pipeline_config_uuid, pipeline_config.feature_configs
        )

        if feature_config_response_status == APIStatus.ERROR:
            return None

        return pipeline_config_uuid

    def _local_hypertuning(
        self,
        hypertune_config: HypertuneConfig,
        data: pd.DataFrame,
        pipeline_config_id: Optional[int] = None,
        model_config: Optional[LinearConfig] = None,
    ) -> List[TrainedModel]:
        """Runs hypertuning locally.

        Args:
            hypertune_config: The config to be used for hypertuning the model.
            data: The raw dataframe to be trained on.
            pipeline_config_id: The id of the pipeline config to be used for training.
            model_config: The config to be used for training the model.

        Returns:
            A list of `TrainedModel` instances.
        """

        prepared_data, _ = self.prepare(data, pipeline_config_id)
        hyperparameter_permutations = list(
            itertools.product(
                hypertune_config.batch_sizes,
                hypertune_config.epochs,
                hypertune_config.learning_rates,
            )
        )
        trained_models = []

        for i, hyperparameters in enumerate(hyperparameter_permutations):
            batch_size, epochs, learning_rate = hyperparameters
            logging.info(
                "Training model %s/%s", i + 1, len(hyperparameter_permutations)
            )

            trained_model = self.train(
                prepared_data.id,
                pipeline_config_id=pipeline_config_id,
                model_config=model_config,
                training_config=TrainingConfig(
                    batch_size=batch_size,
                    epochs=epochs,
                    learning_rate=learning_rate,
                    loss_type=hypertune_config.loss_type,
                ),
            )

            trained_models.append(trained_model)

        return trained_models



"""A Trained Model created for a pipeline."""
from __future__ import annotations

import os
import logging
import pickle
from typing import Optional, Tuple
import numpy as np
import pandas as pd
import torch
from pydantic import Field

from .data import CSVData, replace_missing_values
from .enums import TargetType, APIStatus
from .models import CalibratedLinear
from .types import TrainedModelMetadata
from .api import get_trained_model_metadata, download_trained_model


class TrainedModel(TrainedModelMetadata):
    """A trained calibrated model.

    This model is a container for a trained calibrated model that provides useful
    methods for using the model. The trained calibrated model is the result of running
    the `train` method of a `Pipeline` instance.

    Example:
    ```python
    data = pd.read_csv("data.csv")
    predictions = trained_model.predict(data)
    trained_model.analyze()
    ```

    Attributes:
        dataset_id: The ID of the dataset used to train the model.
        pipeline_uuid: The UUID of the pipeline used to train the model. This will be
            `None` if the trained model has not been analyzed under a pipeline.
        pipeline_config: The configuration of the pipeline used to train the model.
        model_config: The configuration of the model used to train the model.
        training_config: The training configuration used to train the model.
        training_results: The results of training the model.
        model: The trained calibrated model.
        uuid: The UUID of the trained model. This will be `None` if the trained model
            has not been analyzed under a pipeline.
        analysis_url: The URL of the analysis of the trained model. This will be `None`
            if the trained model has not been analyzed under a pipeline.
    """

    model: CalibratedLinear = Field(...)
    allow_model_hosting: bool = False

    class Config:  # pylint: disable=missing-class-docstring,too-few-public-methods
        """Standard Pydantic BaseModel Config."""

        arbitrary_types_allowed = True

    def predict(self, data: pd.DataFrame) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """Returns predictions for the given data.

        Args:
            data: The data to be used for prediction. Must have all columns used for
                training the model to be used.

        Returns:
            A tuple containing an array of predictions and an array of probabilities.
            If the target type is regression, then logits will be `None`. If the target
            type is classification, then the predictions will be logits.
        """
        data = data.loc[:, list(self.pipeline_config.feature_configs.keys())]
        data = replace_missing_values(data, self.pipeline_config.feature_configs)

        csv_data = CSVData(data)
        csv_data.prepare(self.model.features, None)
        inputs = list(csv_data.batch(csv_data.num_examples))[0]
        with torch.no_grad():
            predictions = self.model(inputs).numpy()

        if self.pipeline_config.target_type == TargetType.REGRESSION:
            return predictions, None

        return predictions, 1.0 / (1.0 + np.exp(-predictions))

    def save(self, filepath: str):
        """Saves the trained model to the specified directory.

        Args:
            filepath: The directory to save the trained model to. If the directory does
                not exist, this function will attempt to create it. If the directory
                already exists, this function will overwrite any existing content with
                conflicting filenames.
        """
        if not os.path.exists(filepath):
            os.makedirs(filepath)
        with open(os.path.join(filepath, "trained_model_metadata.pkl"), "wb") as file:
            pickle.dump(self.dict(exclude={"model"}), file)
        model_path = f"{filepath}/trained_ptcm_model.pt"
        torch.save(self.model, model_path)

    @classmethod
    def from_hosted(cls, trained_model_uuid: str) -> TrainedModel:
        """Loads a trained model from the hosted API.

        Args:
            trained_model_uuid: The UUID of the trained model to load.

        Returns:
            A `TrainedModel` instance.
        """

        api_status, metadata = get_trained_model_metadata(trained_model_uuid)
        if api_status == APIStatus.ERROR or metadata.training_results.test_loss is None:
            logging.error(
                "Trained model %s not found. "
                "Model training may still be in progress.",
                trained_model_uuid,
            )
            return None
        downloaded_file_path = download_trained_model(trained_model_uuid)
        model = torch.load(downloaded_file_path)
        model.eval()

        return TrainedModel(**metadata.dict(), model=model)

    @classmethod
    def load(cls, filepath: str) -> TrainedModel:
        """Loads a trained model from the specified filepath.

        Args:
            filepath: The filepath to load the trained model from. The filepath should
                point to a file created by the `save` method of a `TrainedModel`
                instance.

        Returns:
            A `TrainedModel` instance.
        """
        with open(os.path.join(filepath, "trained_model_metadata.pkl"), "rb") as file:
            trained_model_metadata = pickle.load(file)
        model_path = f"{filepath}/trained_ptcm_model.pt"
        model = torch.load(model_path)
        model.eval()

        return TrainedModel(**trained_model_metadata, model=model)



"""Constants used in the SOTAI SDK."""

MISSING_CATEGORY_VALUE = "<Missing Category>"
MISSING_NUMERICAL_VALUE = float("-inf")
INFERENCE_POLLING_INTERVAL = 5
SOTAI_BASE_URL = "https://app.sotai.ai"
SOTAI_API_ENDPOINT = "api/v1"
SOTAI_API_TIMEOUT = 10



"""Tests for Pipeline."""
from unittest.mock import patch
import pandas as pd
import pytest

from sotai import (
    APIStatus,
    DatasetSplit,
    FeatureType,
    HypertuneConfig,
    LossType,
    Metric,
    Pipeline,
    PipelineConfig,
    TargetType,
    NumericalFeatureConfig,
)
from sotai.enums import InferenceConfigStatus
from sotai.constants import SOTAI_BASE_URL


@pytest.mark.parametrize(
    "target_type,expected_primary_metric",
    [(TargetType.CLASSIFICATION, Metric.AUC), (TargetType.REGRESSION, Metric.MSE)],
)
def test_init(
    fixture_feature_names,
    fixture_target,
    target_type,
    expected_primary_metric,
):
    """Tests pipeline initialization for a classification target."""
    pipeline = Pipeline(fixture_feature_names, fixture_target, target_type)
    assert pipeline.name == f"{fixture_target}_{target_type}"
    assert pipeline.target == fixture_target
    assert pipeline.target_type == target_type
    assert pipeline.primary_metric == expected_primary_metric
    assert len(pipeline.feature_configs) == 3
    numerical_config = pipeline.feature_configs["numerical"]
    assert numerical_config.name == "numerical"
    assert numerical_config.type == FeatureType.NUMERICAL
    categorical_strs_config = pipeline.feature_configs["categorical_strs"]
    assert categorical_strs_config.name == "categorical_strs"
    # Note: we expect the default config to be numerical if not specified.
    assert categorical_strs_config.type == FeatureType.NUMERICAL
    categorical_ints_config = pipeline.feature_configs["categorical_ints"]
    assert categorical_ints_config.name == "categorical_ints"
    # Note: we expect the default config to be numerical if not specified.
    assert categorical_ints_config.type == FeatureType.NUMERICAL


def test_init_with_categories(
    fixture_feature_names,
    fixture_target,
    fixture_categories_strs,
    fixture_categories_ints,
):
    """Tests pipeline initialization with specified categories."""
    pipeline = Pipeline(
        fixture_feature_names,
        fixture_target,
        TargetType.CLASSIFICATION,
        categories={
            "categorical_strs": fixture_categories_strs,
            "categorical_ints": fixture_categories_ints,
        },
    )
    categorical_strs_config = pipeline.feature_configs["categorical_strs"]
    assert categorical_strs_config.name == "categorical_strs"
    assert categorical_strs_config.type == FeatureType.CATEGORICAL
    assert categorical_strs_config.categories == fixture_categories_strs
    categorical_ints_config = pipeline.feature_configs["categorical_ints"]
    assert categorical_ints_config.name == "categorical_ints"
    assert categorical_ints_config.type == FeatureType.CATEGORICAL
    assert categorical_ints_config.categories == fixture_categories_ints


@pytest.mark.parametrize(
    "target_type", [TargetType.CLASSIFICATION, TargetType.REGRESSION]
)
@pytest.mark.parametrize("metric", [Metric.AUC, Metric.MSE, Metric.MAE])
@pytest.mark.parametrize("shuffle_data", [True, False])
@pytest.mark.parametrize("drop_empty_percentage", [30, 60, 80])
@pytest.mark.parametrize(
    "dataset_split",
    [
        DatasetSplit(train=80, val=10, test=10),
        DatasetSplit(train=60, val=20, test=20),
        DatasetSplit(train=70, val=20, test=10),
    ],
)
def test_init_from_config(
    fixture_target,
    fixture_feature_configs,
    target_type,
    metric,
    shuffle_data,
    drop_empty_percentage,
    dataset_split,
):
    """Tests pipeline initialization from a `PipelineConfig` instance."""
    pipeline_config = PipelineConfig(
        id=0,
        target=fixture_target,
        target_type=target_type,
        primary_metric=metric,
        feature_configs=fixture_feature_configs,
        shuffle_data=shuffle_data,
        drop_empty_percentage=drop_empty_percentage,
        dataset_split=dataset_split,
    )
    name = "test_pipeline"
    pipeline = Pipeline.from_config(pipeline_config, name=name)
    assert pipeline.name == name
    assert pipeline.target == fixture_target
    assert pipeline.target_type == target_type
    assert pipeline.primary_metric == metric
    assert pipeline.feature_configs == fixture_feature_configs
    assert pipeline.shuffle_data == shuffle_data
    assert pipeline.drop_empty_percentage == drop_empty_percentage
    assert pipeline.dataset_split == dataset_split


def test_prepare(
    fixture_data,
    fixture_feature_names,
    fixture_target,
    fixture_categories_strs,
):
    """Tests the pipeline prepare function."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, target_type=TargetType.CLASSIFICATION
    )
    # We set shuffle to false to ensure the data is split in the same way.
    pipeline.shuffle_data = False
    pipeline.dataset_split.train = 80
    pipeline.dataset_split.val = 10
    pipeline.dataset_split.test = 10
    dataset, pipeline_config = pipeline.prepare(fixture_data)
    assert pipeline_config.id == 0
    categorical_strs_config = pipeline_config.feature_configs["categorical_strs"]
    assert categorical_strs_config.name == "categorical_strs"
    assert categorical_strs_config.type == FeatureType.CATEGORICAL
    assert categorical_strs_config.categories == fixture_categories_strs
    categorical_ints_config = pipeline_config.feature_configs["categorical_ints"]
    assert categorical_ints_config.name == "categorical_ints"
    # Note: integer categories will be detected as numerical if not specified.
    assert categorical_ints_config.type == FeatureType.NUMERICAL
    assert dataset.id == 0
    assert dataset.pipeline_config_id == pipeline_config.id
    num_examples = len(fixture_data)
    num_training_examples = int(num_examples * pipeline.dataset_split.train / 100)
    num_val_examples = int(num_examples * pipeline.dataset_split.val / 100)
    assert dataset.prepared_data.train.equals(fixture_data.iloc[:num_training_examples])
    assert dataset.prepared_data.val.equals(
        fixture_data.iloc[
            num_training_examples : num_training_examples + num_val_examples
        ]
    )
    assert dataset.prepared_data.test.equals(
        fixture_data.iloc[num_training_examples + num_val_examples :]
    )


@pytest.mark.parametrize(
    "target_type",
    [
        (TargetType.CLASSIFICATION),
        (TargetType.REGRESSION),
    ],
)
def test_train_calibrated_linear_model(
    fixture_data,
    fixture_feature_names,
    fixture_target,
    target_type,
):
    """Tests pipeline training for calibrated linear regression model."""
    pipeline = Pipeline(fixture_feature_names, fixture_target, target_type)
    pipeline.shuffle_data = False
    pipeline.dataset_split.train = 60
    pipeline.dataset_split.val = 20
    pipeline.dataset_split.test = 20
    trained_model = pipeline.train(fixture_data)
    assert len(pipeline.configs) == 1
    assert len(pipeline.datasets) == 1
    assert trained_model
    assert trained_model.dataset_id == 0
    assert pipeline.datasets[trained_model.dataset_id]
    assert trained_model.pipeline_config.id == 0
    assert pipeline.configs[trained_model.pipeline_config.id]


def test_pipeline_save_load(
    fixture_data,
    fixture_feature_names,
    fixture_target,
    tmp_path,
):
    """Tests that an instance of `Pipeline` can be successfully saved and loaded."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, TargetType.CLASSIFICATION
    )
    trained_model = pipeline.train(fixture_data)
    pipeline.save(tmp_path)
    loaded_pipeline = Pipeline.load(tmp_path)
    assert isinstance(loaded_pipeline, Pipeline)
    assert loaded_pipeline.name == pipeline.name
    assert loaded_pipeline.target == pipeline.target
    assert loaded_pipeline.target_type == pipeline.target_type
    assert loaded_pipeline.primary_metric == pipeline.primary_metric
    assert loaded_pipeline.feature_configs == pipeline.feature_configs
    assert loaded_pipeline.configs == pipeline.configs
    for dataset_id, loaded_dataset in loaded_pipeline.datasets.items():
        dataset = pipeline.datasets[dataset_id]
        assert loaded_dataset.id == dataset.id
        assert loaded_dataset.pipeline_config_id == dataset.pipeline_config_id
        assert loaded_dataset.prepared_data.train.equals(dataset.prepared_data.train)
        assert loaded_dataset.prepared_data.val.equals(dataset.prepared_data.val)
        assert loaded_dataset.prepared_data.test.equals(dataset.prepared_data.test)
    assert len(loaded_pipeline.trained_models) == 1
    assert loaded_pipeline.trained_models[0].dataset_id == trained_model.dataset_id
    assert (
        loaded_pipeline.trained_models[0].pipeline_uuid == trained_model.pipeline_uuid
    )
    assert (
        loaded_pipeline.trained_models[0].pipeline_config
        == trained_model.pipeline_config
    )
    assert (
        loaded_pipeline.trained_models[0].training_results
        == trained_model.training_results
    )
    assert (
        loaded_pipeline.trained_models[0].training_config
        == trained_model.training_config
    )


@patch(
    "sotai.pipeline.post_pipeline", return_value=(APIStatus.SUCCESS, "test_pipeline_id")
)
def test_publish(
    post_pipeline,
    fixture_feature_names,
    fixture_target,
):
    """Tests that a pipeline can be published to the API."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, TargetType.CLASSIFICATION
    )
    pipeline_uuid = pipeline.publish()
    post_pipeline.assert_called_once()
    assert pipeline_uuid == "test_pipeline_id"


@patch("sotai.pipeline.Pipeline._upload_model", return_value=APIStatus.SUCCESS)
@patch(
    "sotai.pipeline.post_trained_model_analysis",
    return_value=(
        APIStatus.SUCCESS,
        {"trained_model_metadata_uuid": "test_uuid"},
    ),
)
@patch("sotai.pipeline.post_pipeline_feature_configs", return_value=APIStatus.SUCCESS)
@patch(
    "sotai.pipeline.post_pipeline_config",
    return_value=(APIStatus.SUCCESS, "test_pipeline_config_id"),
)
@patch(
    "sotai.pipeline.post_pipeline", return_value=(APIStatus.SUCCESS, "test_pipeline_id")
)
@patch("sotai.pipeline.get_api_key", return_value="test_api_key")
def test_analysis(
    get_api_key,
    post_pipeline,
    post_pipeline_config,
    post_pipeline_feature_configs,
    post_trained_model_analysis,
    upload_model,
    fixture_data,
    fixture_feature_names,
    fixture_target,
):
    """Tests that pipeline analysis works as expected."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, TargetType.CLASSIFICATION
    )
    trained_model = pipeline.train(fixture_data)
    analysis_response = pipeline.analysis(trained_model)

    get_api_key.assert_called()
    upload_model.assert_called_once()
    post_pipeline.assert_called_once()
    post_pipeline_config.assert_called_once_with(
        "test_pipeline_id", trained_model.pipeline_config
    )
    post_pipeline_feature_configs.assert_called_once_with(
        "test_pipeline_config_id", trained_model.pipeline_config.feature_configs
    )
    post_trained_model_analysis.assert_called_once_with(
        "test_pipeline_config_id", trained_model
    )
    expected_analysis_response = (
        f"{SOTAI_BASE_URL}/pipelines/test_pipeline_id/"
        f"trained-models/test_uuid/overall-model-results"
    )
    assert analysis_response == expected_analysis_response


@patch(
    "pandas.read_csv",
    return_value=pd.DataFrame(
        [[1, 2, 3]], columns=["categorical_ints", "numerical", "categorical_strs"]
    ),
)
@patch(
    "sotai.pipeline.post_inference",
    return_value=(APIStatus.SUCCESS, "test_inference_uuid"),
)
@patch("sotai.pipeline.get_api_key", return_value="test_api_key")
def test_run_inference(
    get_api_key,
    post_inference,
    read_csv,
    fixture_data,
    fixture_feature_names,
    fixture_target,
):
    """Tests that a pipeline can run inference on a dataset."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, TargetType.CLASSIFICATION
    )
    trained_model = pipeline.train(fixture_data)
    trained_model.uuid = "test_uuid"

    pipeline.inference("/tmp/data.csv", trained_model)

    read_csv.assert_called_once_with("/tmp/data.csv")
    get_api_key.assert_called_once()
    post_inference.assert_called_once_with("/tmp/data.csv", "test_uuid")


@patch("sotai.pipeline.INFERENCE_POLLING_INTERVAL", 1)
@patch("sotai.pipeline.get_inference_results", response_value=APIStatus.SUCCESS)
@patch(
    "sotai.pipeline.get_inference_status",
    side_effect=[
        (APIStatus.SUCCESS, InferenceConfigStatus.INITIALIZING),
        (APIStatus.SUCCESS, InferenceConfigStatus.SUCCESS),
    ],
)
def test_await_inference_results(
    get_inference_status,
    get_inference_results,
    fixture_data,
    fixture_feature_names,
    fixture_target,
):
    """Tests that a pipeline can await inference results."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, TargetType.CLASSIFICATION
    )
    pipeline.train(fixture_data)

    pipeline.await_inference("test_uuid", "/tmp/predictions")
    get_inference_status.assert_called_with("test_uuid")
    get_inference_results.assert_called_once()


@patch(
    "sotai.pipeline.post_hypertune_job",
    return_value=(
        APIStatus.SUCCESS,
        ["test_uuid"],
    ),
)
@patch(
    "sotai.pipeline.Pipeline._upload_dataset",
    return_value="test_dataset_id",
)
@patch(
    "sotai.pipeline.Pipeline._post_pipeline_config",
    return_value="test_pipeline_config_id",
)
@patch("sotai.pipeline.get_api_key", return_value="test_api_key")
def test_hypertune_hosted(
    get_api_key,
    post_pipeline_config,
    upload_dataset,
    post_hypertune_job,
    fixture_data,
    fixture_feature_names,
    fixture_target,
):
    """Tests that pipeline analysis works as expected."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, TargetType.CLASSIFICATION
    )
    hypertune_config = HypertuneConfig(
        epochs=[100],
        batch_sizes=[32],
        learning_rates=[0.001, 0.01],
        loss_type=LossType.BINARY_CROSSENTROPY,
    )
    hypertune_response = pipeline.hypertune(fixture_data, hypertune_config, hosted=True)

    get_api_key.assert_called()
    upload_dataset.assert_called_once()
    post_hypertune_job.assert_called_once()
    post_pipeline_config.assert_called_once()

    assert hypertune_response[0] == "test_uuid"


def test_hypertune_local(fixture_feature_names, fixture_target, fixture_data):
    """Tests that pipeline hypertuning works as expected."""
    pipeline = Pipeline(
        fixture_feature_names, fixture_target, TargetType.CLASSIFICATION
    )
    hypertune_config = HypertuneConfig(
        epochs=[100],
        batch_sizes=[32, 64],
        learning_rates=[0.001, 0.01],
        loss_type=LossType.BINARY_CROSSENTROPY,
    )
    trained_models = pipeline.hypertune(fixture_data, hypertune_config, hosted=False)

    assert len(trained_models) == 4
    assert trained_models[0].training_config.epochs == 100


@patch(
    "sotai.pipeline.get_pipeline",
    return_value=(
        {
            "id": 1,
            "name": "test_pipeline",
            "target": "target",
            "target_type": "classification",
            "primary_metric": "auc",
        },
        0,
        {
            0: PipelineConfig(
                **{
                    "id": 0,
                    "target": "target",
                    "target_type": "classification",
                    "primary_metric": "auc",
                    "feature_configs": {
                        "age": NumericalFeatureConfig(
                            **{
                                "name": "age",
                                "num_keypoints": 10,
                                "input_keypoints_init": "quantiles",
                                "monotonicity": "none",
                            }
                        ),
                        "chol": NumericalFeatureConfig(
                            **{
                                "name": "chol",
                                "num_keypoints": 10,
                                "input_keypoints_init": "quantiles",
                                "monotonicity": "none",
                            }
                        ),
                    },
                    "shuffle_data": False,
                    "drop_empty_percentage": 70,
                    "dataset_split": {
                        "train": 80,
                        "val": 10,
                        "test": 10,
                    },
                }
            )
        },
        ["test_uuid"],
    ),
)
def test_load_from_hosted(mock_get_pipeline):
    """Tests that a pipeline can be loaded from the API."""
    pipeline = Pipeline.from_hosted("test_uuid", include_trained_models=False)

    mock_get_pipeline.assert_called_once_with("test_uuid")

    assert pipeline.name == "test_pipeline"
    assert pipeline.target == "target"
    assert pipeline.target_type == TargetType.CLASSIFICATION
    assert pipeline.primary_metric == Metric.AUC
    assert len(pipeline.feature_configs) == 2
    assert len(pipeline.configs) == 1
    assert len(pipeline.datasets) == 0


@patch("sotai.pipeline.TrainedModel.from_hosted", return_value=None)
@patch(
    "sotai.pipeline.get_trained_model_uuids",
    return_value=["test_model_uuid"],
)
@patch(
    "sotai.pipeline.Pipeline._upload_model",
    return_value=(APIStatus.SUCCESS),
)
@patch(
    "sotai.pipeline.Pipeline.analysis",
    return_value=(APIStatus.SUCCESS),
)
@patch(
    "sotai.pipeline.Pipeline._upload_dataset",
    return_value=(APIStatus.SUCCESS, "test_dataset_id"),
)
@patch(
    "sotai.pipeline.Pipeline._post_pipeline_config",
    return_value="test_pipeline_config_id",
)
@patch(
    "sotai.pipeline.Pipeline.publish",
    return_value=None,
)
def test_sync(
    publish,
    post_pipeline_config,
    upload_dataset,
    analysis,
    upload_model,
    get_trained_model_uuids,
    from_hosted,
    fixture_data,
    fixture_feature_names,
    fixture_target,
):
    """Tests that a pipeline can be synced."""
    pipeline = Pipeline(
        fixture_feature_names,
        fixture_target,
        TargetType.CLASSIFICATION,
        default_allow_hosting=True,
    )

    pipeline.hypertune(
        fixture_data,
        hypertune_config=HypertuneConfig(
            epochs=[100, 200],
            batch_sizes=[32],
            learning_rates=[0.01],
            loss_type=LossType.BINARY_CROSSENTROPY,
        ),
        hosted=False,
    )
    pipeline.sync()

    publish.assert_called_once()
    from_hosted.assert_called_once()
    upload_dataset.assert_called_once()
    post_pipeline_config.assert_called_once()
    analysis.assert_called()
    upload_model.assert_called()
    get_trained_model_uuids.assert_called_once()



"""Tests for configuration objects."""
import numpy as np
import pytest

from sotai import FeatureType, InputKeypointsInit, Monotonicity
from sotai.features import CategoricalFeature, NumericalFeature


@pytest.mark.parametrize(
    "data,num_keypoints,input_keypoints_init,missing_input_value,monotonicity,"
    "expected_input_keypoints",
    [
        (
            np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
            5,
            InputKeypointsInit.QUANTILES,
            None,
            Monotonicity.NONE,
            np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
        ),
        (
            np.array([1.0] * 10 + [2.0] * 10 + [3.0] * 10 + [4.0] * 10 + [5.0] * 10),
            5,
            InputKeypointsInit.QUANTILES,
            None,
            Monotonicity.INCREASING,
            np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
        ),
        (
            np.array([1.0] * 10 + [2.0] * 10 + [3.0] * 10 + [4.0] * 10 + [5.0] * 10),
            5,
            InputKeypointsInit.UNIFORM,
            None,
            Monotonicity.INCREASING,
            np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
        ),
        (
            np.array(
                [1.0] * 10 + [2.0] * 8 + [3.0] * 6 + [4.0] * 4 + [5.0] * 2 + [6.0]
            ),
            4,
            InputKeypointsInit.QUANTILES,
            None,
            Monotonicity.INCREASING,
            np.array([1.0, 3.0, 4.0, 6.0]),
        ),
        (
            np.array(
                [1.0] * 10 + [2.0] * 8 + [3.0] * 6 + [4.0] * 4 + [5.0] * 2 + [6.0]
            ),
            10,
            InputKeypointsInit.QUANTILES,
            None,
            Monotonicity.NONE,
            np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]),
        ),
    ],
)
def test_numerical_feature_config_initialization(
    data,
    num_keypoints,
    input_keypoints_init,
    missing_input_value,
    monotonicity,
    expected_input_keypoints,
) -> None:
    """Tests that numerical feature configs initialize properly."""
    feature_name = "test_feature"
    feature = NumericalFeature(
        feature_name,
        data,
        num_keypoints,
        input_keypoints_init,
        missing_input_value,
        monotonicity,
    )
    assert feature.feature_type == FeatureType.NUMERICAL
    assert feature.feature_name == feature_name
    assert (feature.data == data).all()
    assert feature.num_keypoints == num_keypoints
    assert feature.input_keypoints_init == input_keypoints_init
    assert feature.missing_input_value == missing_input_value
    assert feature.monotonicity == monotonicity
    assert np.allclose(feature.input_keypoints, expected_input_keypoints)


@pytest.mark.parametrize(
    "categories,missing_input_value,monotonicity_pairs,expected_category_indices,"
    "expected_monotonicity_index_pairs",
    [
        (["a", "b", "c"], None, None, {"a": 0, "b": 1, "c": 2}, []),
        (
            ["a", "b", "c", "d"],
            -1.0,
            [("a", "b"), ("c", "d")],
            {"a": 0, "b": 1, "c": 2, "d": 3},
            [(0, 1), (2, 3)],
        ),
    ],
)
def test_categorical_feature_config_initialization(
    categories,
    missing_input_value,
    monotonicity_pairs,
    expected_category_indices,
    expected_monotonicity_index_pairs,
):
    """Tests that categorical feature configs initialize properly."""
    feature_name = "test_feature"
    feature = CategoricalFeature(
        feature_name, categories, missing_input_value, monotonicity_pairs
    )
    assert feature.feature_type == FeatureType.CATEGORICAL
    assert feature.feature_name == feature_name
    assert feature.categories == categories
    assert feature.missing_input_value == missing_input_value
    assert feature.monotonicity_pairs == monotonicity_pairs
    assert feature.category_indices == expected_category_indices
    assert feature.monotonicity_index_pairs == expected_monotonicity_index_pairs



"""Linear module for use in calibrated modeling.

PyTorch implementation of the calibrated linear module. This module takes in a
single-dimensional input and transforms it using a linear transformation and optionally
a bias term. This module supports monotonicity constraints.
"""
from typing import List, Optional

import torch

from ..enums import Monotonicity


class Linear(torch.nn.Module):
    """A linear module.

    This module takes an input of shape `(batch_size, input_dim)` and applied a linear
    transformation. The output will have the same shape as the input.

    Attributes:
      - All `__init__` arguments.
      kernel: `torch.nn.Parameter` that stores the linear combination weighting.
      bias: `torch.nn.Parameter` that stores the bias term. Only available is `use_bias`
        is true.

    Example:
    ```python
    input_dim = 3
    inputs = torch.tensor(...)  # shape: (batch_size, input_dim)
    linear = Linear(
      input_dim,
      monotonicities=[
        Monotonicity.NONE,
        Monotonicity.INCREASING,
        Monotonicity.DECREASING
      ],
      use_bias=False,
      weighted_average=True,
    )
    outputs = linear(inputs)
    ```
    """

    def __init__(
        self,
        input_dim,
        monotonicities: Optional[List[Monotonicity]] = None,
        use_bias: bool = True,
        weighted_average: bool = False,
    ) -> None:
        """Initializes an instance of `Linear`.

        Args:
            input_dim: The number of inputs that will be combined.
            monotonicities: If provided, specifies the monotonicity of each input
              dimension.
            use_bias: Whether to use a bias term for the linear combination.
            weighted_average: Whether to make the output a weighted average i.e. all
              coefficients are positive and add up to a total of 1.0. No bias term will
              be used, and `use_bias` will be set to false regardless of the original
              value. `monotonicities` will also be set to increasing for all input
              dimensions to ensure that all coefficients are positive.

        Raises:
            ValueError: If monotonicities does not have length input_dim (if provided).
        """
        super().__init__()

        self.input_dim = input_dim
        if monotonicities and len(monotonicities) != input_dim:
            raise ValueError("Monotonicities, if provided, must have length input_dim.")
        self.monotonicities = (
            monotonicities
            if not weighted_average
            else [Monotonicity.INCREASING] * input_dim
        )
        self.use_bias = use_bias if not weighted_average else False
        self.weighted_average = weighted_average

        self.kernel = torch.nn.Parameter(torch.Tensor(input_dim, 1).double())
        torch.nn.init.constant_(self.kernel, 1.0 / input_dim)
        if use_bias:
            self.bias = torch.nn.Parameter(torch.Tensor(1).double())
            torch.nn.init.constant_(self.bias, 0.0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # pylint: disable=invalid-name
        """Transforms inputs using a linear combination.

        Args:
            x: The input tensor of shape `(batch_size, input_dim)`.

        Returns:
            torch.Tensor of shape `(batch_size, 1)` containing transformed input values.
        """
        result = torch.mm(x, self.kernel)
        if self.use_bias:
            result += self.bias
        return result

    @torch.no_grad()
    def assert_constraints(self, eps=1e-6) -> List[str]:
        """Asserts that layer satisfies specified constraints.

        This checks that decreasing monotonicity corresponds to negative weights,
        increasing monotonicity corresponds to positive weights, and weights sum to 1
        for weighted_average=True.

        Args:
            eps: the margin of error allowed

        Returns:
            A list of messages describing violated constraints. If no constraints
            violated, the list will be empty.
        """
        messages = []

        if self.weighted_average:
            total_weight = torch.sum(self.kernel.data)
            if torch.abs(total_weight - 1.0) > eps:
                messages.append("Weights do not sum to 1.")

        if self.monotonicities:
            monotonicities_constant = torch.tensor(
                [
                    1
                    if m == Monotonicity.INCREASING
                    else -1
                    if m == Monotonicity.DECREASING
                    else 0
                    for m in self.monotonicities
                ],
                device=self.kernel.device,
                dtype=self.kernel.dtype,
            ).view(-1, 1)

            violated_monotonicities = (self.kernel * monotonicities_constant) < -eps
            violation_indices = torch.where(violated_monotonicities)
            if violation_indices[0].numel() > 0:
                messages.append(
                    f"Monotonicity violated at: {violation_indices[0].tolist()}"
                )

        return messages

    @torch.no_grad()
    def constrain(self) -> None:
        """Projects kernel into desired constraints."""
        projected_kernel_data = self.kernel.data

        if self.monotonicities:
            if Monotonicity.INCREASING in self.monotonicities:
                increasing_mask = torch.tensor(
                    [
                        [0.0] if m == Monotonicity.INCREASING else [1.0]
                        for m in self.monotonicities
                    ]
                )
                projected_kernel_data = torch.maximum(
                    projected_kernel_data, projected_kernel_data * increasing_mask
                )
            if Monotonicity.DECREASING in self.monotonicities:
                decreasing_mask = torch.tensor(
                    [
                        [0.0] if m == Monotonicity.DECREASING else [1.0]
                        for m in self.monotonicities
                    ]
                )
                projected_kernel_data = torch.minimum(
                    projected_kernel_data, projected_kernel_data * decreasing_mask
                )

        if self.weighted_average:
            norm = torch.norm(projected_kernel_data, 1)
            norm = torch.where(norm < 1e-8, 1.0, norm)
            projected_kernel_data /= norm

        self.kernel.data = projected_kernel_data



"""Useful classes and functions for handling data for calibrated modeling."""
import logging
from typing import Dict, Generator, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch

from .constants import MISSING_CATEGORY_VALUE, MISSING_NUMERICAL_VALUE
from .enums import FeatureType
from .features import CategoricalFeature, NumericalFeature
from .types import CategoricalFeatureConfig, NumericalFeatureConfig


def replace_missing_values(
    data: pd.DataFrame,
    feature_configs: Dict[str, Union[CategoricalFeatureConfig, NumericalFeatureConfig]],
) -> pd.DataFrame:
    """Replaces empty values or unspecified categories with a constant value.

    Args:
        - data: The dataset in which to replace missing values.
        - feature_configs: A dictionary mapping feature names to feature configurations.

    Returns:
        The dataset with missing values replaced.
    """
    for feature_name, feature_config in feature_configs.items():
        if feature_config.type == FeatureType.CATEGORICAL:
            unspecified_categories = list(
                set(data[feature_name].unique().tolist())
                - set(feature_config.categories)
            )
            if unspecified_categories:
                logging.info(
                    "Replacing %s with %s for feature %s",
                    unspecified_categories,
                    MISSING_CATEGORY_VALUE,
                    feature_name,
                )
                data[feature_name].replace(
                    unspecified_categories, MISSING_CATEGORY_VALUE, inplace=True
                )
            data[feature_name].fillna(MISSING_CATEGORY_VALUE, inplace=True)
        else:
            data[feature_name].fillna(MISSING_NUMERICAL_VALUE, inplace=True)

    return data


def determine_feature_types(data: pd.DataFrame) -> Dict[str, FeatureType]:
    """Returns a dictionary mapping feature name to type for the given data."""
    feature_types = {}
    for column in data.columns:
        dtype_kind = data[column].dtype.kind
        if dtype_kind in ["S", "O", "b"]:  # string, object, boolean
            feature_types[column] = FeatureType.CATEGORICAL
        elif dtype_kind in ["i", "u", "f"]:  # integer, unsigned integer, float
            feature_types[column] = FeatureType.NUMERICAL
        else:  # datetime, timedelta, complex, etc.
            feature_types[column] = FeatureType.UNKNOWN
    return feature_types


class CSVData:
    """Class for handling CSV data for calibrated modeling.

    Attributes:
        - All `__init__` arguments.
        data: A pandas `DataFrame` containing the loaded CSV data.
        headers: The list of headers available from the loaded data.
        num_examples: The number of examples in the dataset.
        prepared_data: The prepared data. This will be `None` if `prepare(...)` has not
            been called.

    Example:
    ```python
    csv_data = CSVData("path/to/data.csv")
    feature_configs = [
        NumericalFeatureConfig(
            feature_name="numerical_feature"
            data=csv_data("numerical_feature")  # must match feature column header
        ),
        CategoricalFeatureConfig(
            feature_name="categorical_feature"
            categories=np.unique(csv_data("categorical_feature"))  # uses all categories
        ),
    ]
    csv_data.prepare(feature_configs, "target", ...)  # must match target column header
    for examples, labels in csv_data.batch(64):
        training_step(...)
    ```
    """

    def __init__(
        self,
        dataset: Union[str, pd.DataFrame],
    ) -> None:
        """Initializes an instance of `CSVData`.

        Loads a CSV file if filepath is provided. Otherwise it will use the provided
        DataFrame.

        Args:
            dataset: Either a string filepath pointing to the CSV data that should be
                loaded or a `pd.DataFrame` containing the data that should be used.
                The CSV file or DataFrame must have a header.
        """
        self.dataset = dataset
        if isinstance(dataset, str):
            self.data = pd.read_csv(dataset)
        else:
            self.data = dataset.copy()
        self.headers = list(self.data.columns)
        self.num_examples = len(self.data)
        self.prepared_data = None
        self._prepared_data_tensor, self._targets_tensor = None, None

    def __call__(self, header: str, prepared: bool = False) -> np.ndarray:
        """Selects the data for the column with the given header.

        Args:
            header: The header of the column for which data is selected.

        Returns:
            The selected data as a numpy array.

        Raises:
            ValueError: If the provided header is not available in the data.
            ValueError: If `prepared` is true but `prepare(...)` has not been called.
        """
        if not header in self.headers:
            raise ValueError(f"Column {header} does not exist: {self.headers}.")
        if prepared:
            if self.prepared_data is None:
                raise ValueError(
                    "CSV.prepare(...) must be called first for prepared data."
                )
            return self.prepared_data[header].to_numpy()
        return self.data[header].to_numpy()

    def prepare(
        self,
        features: List[Union[NumericalFeature, CategoricalFeature]],
        target_header: Optional[str],
        inplace: bool = True,
    ) -> None:
        """Prepares the data for calibrated modeling.

        Args:
            feature_configs: Feature configs that specify how to prepare the data.
            target_header: The header for the target column. If `None`, it will be
                assumed that there is no target column present (e.g. for inference)
            inplace: If True, original `data` attribute will be updated. If False, a
                copy of the original data will be prepared and the original will be
                preserved.

        Raises:
            ValueError: If a feature in `feature_configs` is not in the dataset.
        """
        if self.prepared_data is not None:
            logging.info("Prepare has already been called. Doing nothing.")
            return
        selected_features = [feature.feature_name for feature in features]
        unavailable_features = set(selected_features) - set(self.data.columns)
        if len(unavailable_features) > 0:
            raise ValueError(f"Features {unavailable_features} not found in dataset.")
        self.prepared_data = self.data if inplace else self.data.copy()
        for feature in features:
            if feature.feature_type == FeatureType.CATEGORICAL:
                feature_data = self.prepared_data[feature.feature_name].map(
                    feature.category_indices
                )
                if feature.missing_input_value is not None:
                    feature_data = feature_data.fillna(feature.missing_input_value)
                self.prepared_data[feature.feature_name] = feature_data
        if target_header is not None:
            self._targets_tensor = torch.from_numpy(
                self.prepared_data.pop(target_header).values
            )[:, None].double()
        self._prepared_data_tensor = torch.from_numpy(
            self.prepared_data[selected_features].values
        ).double()

    def batch(
        self, batch_size: int
    ) -> Generator[Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor], None, None]:
        """A generator that yields a tensor with `batch_size` examples.

        Args:
            batch_size: The size of each batch returns during each iteration.

        Yields:
            If prepared with a target column: a tuple (examples, targets) of
                `torch.Tensor` of shape `(batch_size, num_features)` and
                `(batch_size, 1)`, repsectively.
            If prepared without a target column: a `torch.Tensor` of shape
                `(batch_size, num_features)`.

        Raises:
            ValueError: If `prepare(...)` is not called first.
        """
        if self.prepared_data is None:
            raise ValueError("CSVData.prepare(...) must be called first.")

        for i in range(0, self.num_examples, batch_size):
            if self._targets_tensor is not None:
                yield (
                    self._prepared_data_tensor[i : i + batch_size],
                    self._targets_tensor[i : i + batch_size],
                )
            else:
                yield self._prepared_data_tensor[i : i + batch_size]



"""Functions for quickly prepping demo data to use with SOTAI."""
from typing import List, Tuple

import pandas as pd


def heart() -> Tuple[pd.DataFrame, List[str], str]:
    """Prepares the demo heart dataset for use with the SOTAI Quickstart guide.

    The heart dataset is a classification dataset with 303 rows and 14 columns. The
    target is binary, with 0 indicating no heart disease and 1 indicating heart
    disease. The features are a mix of categorical and numerical features. For more
    information, see https://archive.ics.uci.edu/ml/datasets/heart+Disease.

    Returns:
        A tuple containing the cleaned heart dataset as a pandas `DataFrame`, the list
        of features, and the target.
    """
    heart_data = pd.read_csv(
        "http://storage.googleapis.com/download.tensorflow.org/data/heart.csv"
    )
    features = list(heart_data.columns)
    target = features.pop(features.index("target"))
    # Convert numerical categories to strings so they are handled as categories
    categorical_features = ["sex", "fbs", "restecg", "exang", "thal"]
    heart_data[categorical_features] = heart_data[categorical_features].astype(str)
    # Clean the thal feature categories
    heart_data["thal"].replace({"1": "fixed", "2": "reversible"}, inplace=True)
    return heart_data, features, target



"""Enum Classes for SOTAI SDK."""
from enum import Enum, EnumMeta
from typing import Any


class _Metaclass(EnumMeta):
    """Base `EnumMeta` subclass for accessing enum members directly."""

    def __getattribute__(cls, __name: str) -> Any:
        value = super().__getattribute__(__name)
        if isinstance(value, Enum):
            value = value.value
        return value


class _Enum(str, Enum, metaclass=_Metaclass):
    """Base Enum Class."""


class TargetType(_Enum):
    """The type of target to predict.

    - CLASSIFICATION: classification target i.e. binary 0/1.
    - REGRESSION: regression target i.e. continuous float.
    """

    CLASSIFICATION = "classification"
    REGRESSION = "regression"


class LossType(_Enum):
    """The type of loss function to use.

    - BINARY_CROSSENTROPY: binary cross entropy loss.
    - HINGE: hinge loss.
    - HUBER: huber loss.
    - MAE: mean absolute error loss.
    - MSE: mean squared error loss.
    """

    BINARY_CROSSENTROPY = "binary"
    HINGE = "hinge"
    HUBER = "huber"
    MAE = "mae"
    MSE = "mse"


class Metric(_Enum):
    """The type of metric to use.

    - AUC: area under the ROC curve.
    - MAE: mean absolute error.
    - MSE: mean squared error.
    """

    AUC = "auc"
    MAE = "mean_absolute_error"
    MSE = "mean_squared_error"


class InputKeypointsInit(_Enum):
    """Type of initialization to use for NumericalCalibrator input keypoints.

    - QUANTILES: initialize the input keypoints such that each segment will see the same
        number of examples.
    - UNIFORM: initialize the input keypoints uniformly spaced in the feature range.
    """

    QUANTILES = "quantiles"
    UNIFORM = "uniform"


class InputKeypointsType(_Enum):
    """The type of input keypoints to use.

    - FIXED: the input keypoints will be fixed during initialization.
    """

    FIXED = "fixed"
    # TODO: add learned interior functionality
    # LEARNED = "learned_interior"


class FeatureType(_Enum):
    """Type of feature.

    - UNKNOWN: a feature with a type that our system does not currently support.
    - NUMERICAL: a numerical feature that should be calibrated using an instance of
        `NumericalCalibrator`.
    - CATEGORICAL: a categorical feature that should be calibrated using an instance of
        `CategoricalCalibrator`.
    """

    UNKNOWN = "unknown"
    NUMERICAL = "numerical"
    CATEGORICAL = "categorical"


class NumericalCalibratorInit(_Enum):
    """Type of kernel initialization to use for NumericalCalibrator.

    - EQUAL_HEIGHTS: initialize the kernel such that all segments have the same height.
    - EQUAL_SLOPES: initialize the kernel such that all segments have the same slope.
    """

    EQUAL_HEIGHTS = "equal_heights"
    EQUAL_SLOPES = "equal_slopes"


class CategoricalCalibratorInit(_Enum):
    """Type of kernel initialization to use for CategoricalCalibrator.

    - UNIFORM: initialize the kernel with uniformly distributed values. The sample range
        will be [`output_min`, `output_max`] if both are provided.
    - CONSTANT: initialize the kernel with a constant value for all categories. This
        value will be `(output_min + output_max) / 2` if both are provided.
    """

    UNIFORM = "uniform"
    CONSTANT = "constant"


class Monotonicity(_Enum):
    """Type of monotonicity constraint.

    - NONE: no monotonicity constraint.
    - INCREASING: increasing monotonicity i.e. increasing input increases output.
    - DECREASING: decreasing monotonicity i.e. increasing input decreases output.
    """

    NONE = "none"
    INCREASING = "increasing"
    DECREASING = "decreasing"


class APIStatus(_Enum):
    """Status of API.

    - SUCCESS: API call was successful
    - ERROR: API call was unsuccessful
    """

    SUCCESS = "success"
    ERROR = "error"


class InferenceConfigStatus(_Enum):
    """Enum for InferenceConfig status.

    - FAILED: inference job failed.
    - INITIALIZING: inference job is initializing.
    - PREPARING: inference job is preparing to run.
    - RUNNING: inference job is running.
    - SUCCESS: inference job completed successfully.
    """

    FAILED = "failed"
    INITIALIZING = "initializing"
    PREPARING = "preparing"
    RUNNING = "running"
    SUCCESS = "success"


class HypertuneMethod(_Enum):
    """Type of hyperparameter optimzation method.

    - GRID: searches the grid generated by taking the cross product of all
        hyperparameter options.
    """

    GRID = "grid"


class Interpolation(_Enum):
    """Enum for interpolation method of lattice.

    - HYPERCUBE: n-dimensional hypercube surrounding input point(s).
    - SIMPLEX: uses only one of the n! simplices in the n-dim hypercube.
    """

    HYPERCUBE = "hypercube"
    SIMPLEX = "simplex"


class LatticeInit(_Enum):
    """Type of kernel initialization to use for CategoricalCalibrator.

    - LINEAR: initialize the kernel with weights represented by a linear function,
        conforming to monotonicity and unimodality constraints.
    - RANDOM_MONOTONIC: initialize the kernel with a uniformly random sampled
        lattice layer weight tensor, conforming to monotonicity and unimodality
        constraints.
    """

    LINEAR = "linear"
    RANDOM_MONOTONIC = "random_monotonic"



"""SOTAI SDK."""

# This version must always be one version ahead of the current release, so it
# matches the current state of development, which will always be ahead of the
# current release. Use Semantic Versioning.
__version__ = "0.6.3"

from . import api, constants, data, demo, external, features, layers, models, training
from .enums import *
from .pipeline import *
from .trained_model import *
from .types import *



"""Tests for models."""
import numpy as np
import pytest
import torch

from sotai import Monotonicity
from sotai.features import CategoricalFeature, NumericalFeature
from sotai.models import CalibratedLinear

from .utils import train_calibrated_module


@pytest.mark.parametrize(
    "features,output_min,output_max,output_calibration_num_keypoints,"
    "expected_linear_monotonicities,expected_output_calibrator_monotonicity",
    [
        (
            [
                NumericalFeature(
                    feature_name="numerical_feature",
                    data=np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
                    num_keypoints=5,
                    monotonicity=Monotonicity.NONE,
                ),
                CategoricalFeature(
                    feature_name="categorical_feature",
                    categories=["a", "b", "c"],
                    monotonicity_pairs=[("a", "b")],
                ),
            ],
            None,
            None,
            None,
            [
                Monotonicity.NONE,
                Monotonicity.INCREASING,
            ],
            Monotonicity.INCREASING,
        ),
        (
            [
                NumericalFeature(
                    feature_name="numerical_feature",
                    data=np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
                    num_keypoints=5,
                    monotonicity=Monotonicity.NONE,
                ),
                CategoricalFeature(
                    feature_name="categorical_feature",
                    categories=["a", "b", "c"],
                    monotonicity_pairs=None,
                ),
            ],
            -1.0,
            1.0,
            10,
            [
                Monotonicity.INCREASING,
                Monotonicity.INCREASING,
            ],
            Monotonicity.NONE,
        ),
        (
            [
                NumericalFeature(
                    feature_name="numerical_feature",
                    data=np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
                    num_keypoints=5,
                    monotonicity=Monotonicity.DECREASING,
                ),
                CategoricalFeature(
                    feature_name="categorical_feature",
                    categories=["a", "b", "c"],
                    monotonicity_pairs=None,
                ),
            ],
            0.0,
            None,
            None,
            [
                Monotonicity.INCREASING,
                Monotonicity.INCREASING,
            ],
            Monotonicity.INCREASING,
        ),
    ],
)
def test_initialization(
    features,
    output_min,
    output_max,
    output_calibration_num_keypoints,
    expected_linear_monotonicities,
    expected_output_calibrator_monotonicity,
):
    """Tests that `CalibratedLinear` initialization works."""
    calibrated_linear = CalibratedLinear(
        features=features,
        output_min=output_min,
        output_max=output_max,
        output_calibration_num_keypoints=output_calibration_num_keypoints,
    )
    assert calibrated_linear.features == features
    assert calibrated_linear.output_min == output_min
    assert calibrated_linear.output_max == output_max
    assert (
        calibrated_linear.output_calibration_num_keypoints
        == output_calibration_num_keypoints
    )
    assert len(calibrated_linear.calibrators) == len(features)
    for calibrator in calibrated_linear.calibrators.values():
        assert calibrator.output_min == output_min
        assert calibrator.output_max == output_max
    assert calibrated_linear.linear.monotonicities == expected_linear_monotonicities
    if (
        output_min is not None
        or output_max is not None
        or output_calibration_num_keypoints
    ):
        assert not calibrated_linear.linear.use_bias
        assert calibrated_linear.linear.weighted_average
    else:
        assert calibrated_linear.linear.use_bias
        assert not calibrated_linear.linear.weighted_average
    if not output_calibration_num_keypoints:
        assert calibrated_linear.output_calibrator is None
    else:
        assert calibrated_linear.output_calibrator.output_min == output_min
        assert calibrated_linear.output_calibrator.output_max == output_max
        assert (
            calibrated_linear.output_calibrator.monotonicity
            == expected_output_calibrator_monotonicity
        )


# TODO: add more parameterized tests
@pytest.mark.parametrize(
    "output_min,output_max,calibrator_kernel_datas,linear_kernel_data,"
    "output_calibrator_kernel_data,inputs,expected_outputs",
    [
        (
            None,
            None,
            [
                torch.tensor([[0.0], [1.0], [1.0], [1.0]]).double(),
                torch.tensor([[1.0], [2.0], [3.0]]).double(),
            ],
            torch.tensor([[1.0], [2.0]]).double(),
            None,
            torch.tensor([[1.0, 0], [2.0, 1], [3.0, 2], [4.0, 1]]).double(),
            torch.tensor([[2.0], [5.0], [8.0], [7.0]]).double(),
        ),
        (
            0.0,
            1.0,
            [
                torch.tensor([[1.0], [-0.5], [-0.5], [1.0]]).double(),
                torch.tensor([[1.0], [0.5], [0.0]]).double(),
            ],
            torch.tensor([[0.3], [0.7]]).double(),
            torch.tensor([[0.0], [1.0], [-1.0], [1.0]]).double(),
            torch.tensor([[1.0, 0], [2.0, 1], [3.0, 2], [4.0, 1]]).double(),
            torch.tensor([[1.0], [0.5], [0.0], [0.05]]).double(),
        ),
    ],
)
def test_forward(
    output_min,
    output_max,
    calibrator_kernel_datas,
    linear_kernel_data,
    output_calibrator_kernel_data,
    inputs,
    expected_outputs,
):
    """Tests that forward returns expected result."""
    calibrated_linear = CalibratedLinear(
        features=[
            NumericalFeature(
                feature_name="numerical_feature",
                data=np.array([1.0, 2.0, 3.0, 4.0]),
                num_keypoints=4,
                monotonicity=Monotonicity.NONE,
            ),
            CategoricalFeature(
                feature_name="categorical_feature",
                categories=["a", "b", "c"],
                monotonicity_pairs=None,
            ),
        ],
        output_min=output_min,
        output_max=output_max,
        output_calibration_num_keypoints=output_calibrator_kernel_data.size()[0]
        if output_calibrator_kernel_data is not None
        else None,
    )
    for i, calibrator in enumerate(calibrated_linear.calibrators.values()):
        calibrator.kernel.data = calibrator_kernel_datas[i]
    calibrated_linear.linear.kernel.data = linear_kernel_data
    if output_calibrator_kernel_data is not None:
        calibrated_linear.output_calibrator.kernel.data = output_calibrator_kernel_data
    outputs = calibrated_linear(inputs)
    assert torch.allclose(outputs, expected_outputs)


@pytest.mark.parametrize(
    "cat_cal_kernel_data,num_cal_kernel_data,linear_kernel_data,weighted_avg,expected_outputs",
    [
        (
            torch.tensor([[2.0], [1.0]]).double(),
            torch.tensor([[1.0], [2.0]]).double(),
            torch.tensor([[0.4], [0.6]]).double(),
            False,
            {"categorical_feature": ["Monotonicity violated at: [(0, 1)]."]},
        ),
        (
            torch.tensor([[1.0], [2.0]]).double(),
            torch.tensor([[2.0], [1.0]]).double(),
            torch.tensor([[0.4], [0.6]]).double(),
            False,
            {"numerical_feature": ["Monotonicity violated at: [(0, 1)]."]},
        ),
        (
            torch.tensor([[1.0], [2.0]]).double(),
            torch.tensor([[1.0], [2.0]]).double(),
            torch.tensor([[0.4], [0.4]]).double(),
            True,
            {"linear": ["Weights do not sum to 1."]},
        ),
        (
            torch.tensor([[2.0], [1.0]]).double(),
            torch.tensor([[2.0], [1.0]]).double(),
            torch.tensor([[0.2], [-0.2]]).double(),
            True,
            {
                "numerical_feature": ["Monotonicity violated at: [(0, 1)]."],
                "categorical_feature": ["Monotonicity violated at: [(0, 1)]."],
                "linear": ["Weights do not sum to 1.", "Monotonicity violated at: [1]"],
            },
        ),
        (
            torch.tensor([[1.0], [2.0]]).double(),
            torch.tensor([[1.0], [2.0]]).double(),
            torch.tensor([[0.4], [0.6]]).double(),
            True,
            {},
        ),
    ],
)
def test_assert_constraints(
    cat_cal_kernel_data,
    num_cal_kernel_data,
    linear_kernel_data,
    weighted_avg,
    expected_outputs,
):
    """
    Tests that each layer's assert_constraints is properly called and aggregates each
    set of error messages properly.
    """
    calibrated_linear = CalibratedLinear(
        features=[
            NumericalFeature(
                feature_name="numerical_feature",
                data=np.array([1.0, 2.0]),
                num_keypoints=2,
                monotonicity=Monotonicity.INCREASING,
            ),
            CategoricalFeature(
                feature_name="categorical_feature",
                categories=["a", "b"],
                monotonicity_pairs=[("a", "b")],
            ),
        ],
    )
    calibrated_linear.calibrators[
        "categorical_feature"
    ].kernel.data = cat_cal_kernel_data
    calibrated_linear.calibrators["numerical_feature"].kernel.data = num_cal_kernel_data
    calibrated_linear.linear.kernel.data = linear_kernel_data
    calibrated_linear.linear.weighted_average = weighted_avg

    assert calibrated_linear.assert_constraints() == expected_outputs


def test_constrain():
    """Tests that constrain properly constrains all layers."""
    output_min, output_max = -1.0, 1.0
    calibrated_linear = CalibratedLinear(
        features=[
            NumericalFeature(
                feature_name="numerical_feature",
                data=np.array([1.0, 2.0, 3.0, 4.0]),
                num_keypoints=4,
                monotonicity=Monotonicity.NONE,
            ),
            CategoricalFeature(
                feature_name="categorical_feature",
                categories=["a", "b", "c"],
                monotonicity_pairs=None,
            ),
        ],
        output_min=-output_min,
        output_max=output_max,
        output_calibration_num_keypoints=5,
    )
    calibrated_linear.calibrators["numerical_feature"].kernel.data = torch.tensor(
        [[-2.0], [1.0], [2.0], [1.0]]
    ).double()
    calibrated_linear.calibrators["categorical_feature"].kernel.data = torch.tensor(
        [[2.0], [0.5], [-2.0]]
    ).double()
    calibrated_linear.linear.kernel.data = torch.tensor([[1.0], [2.0]]).double()
    calibrated_linear.output_calibrator.kernel.data = torch.tensor(
        [[0.0], [1.0], [0.5], [1.0], [-1.0]]
    ).double()
    calibrated_linear.constrain()
    for calibrator in calibrated_linear.calibrators.values():
        keypoint_outputs = calibrator.keypoints_outputs()
        assert torch.all(keypoint_outputs >= output_min)
        assert torch.all(keypoint_outputs <= output_max)
    assert torch.sum(calibrated_linear.linear.kernel.data) == 1.0
    output_calibrator_keypoint_outputs = (
        calibrated_linear.output_calibrator.keypoints_outputs()
    )
    assert torch.all(output_calibrator_keypoint_outputs >= output_min)
    assert torch.all(output_calibrator_keypoint_outputs <= output_max)


def test_training():  # pylint: disable=too-many-locals
    """Tests `CalibratedLinear` training on data from f(x) = 0.7|x_1| + 0.3x_2."""
    num_examples, num_categories = 3000, 3
    output_min, output_max = 0.0, num_categories - 1
    x_1_numpy = np.random.uniform(-output_max, output_max, size=num_examples)
    x_1 = torch.from_numpy(x_1_numpy)[:, None]
    num_examples_per_category = num_examples // num_categories
    x2_numpy = np.concatenate(
        [[c] * num_examples_per_category for c in range(num_categories)]
    )
    x_2 = torch.from_numpy(x2_numpy)[:, None]
    training_examples = torch.column_stack((x_1, x_2))
    linear_coefficients = torch.tensor([0.7, 0.3]).double()
    training_labels = torch.sum(
        torch.column_stack((torch.absolute(x_1), x_2)) * linear_coefficients,
        dim=1,
        keepdim=True,
    )
    randperm = torch.randperm(training_examples.size()[0])
    training_examples = training_examples[randperm]
    training_labels = training_labels[randperm]

    calibrated_linear = CalibratedLinear(
        features=[
            NumericalFeature(
                "x1",
                x_1_numpy,
                num_keypoints=4,
            ),
            CategoricalFeature("x2", [0, 1, 2], monotonicity_pairs=[(0, 1), (1, 2)]),
        ],
        output_min=output_min,
        output_max=output_max,
    )

    loss_fn = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(calibrated_linear.parameters(recurse=True), lr=1e-1)

    with torch.no_grad():
        initial_predictions = calibrated_linear(training_examples)
        initial_loss = loss_fn(initial_predictions, training_labels)

    train_calibrated_module(
        calibrated_linear,
        training_examples,
        training_labels,
        loss_fn,
        optimizer,
        500,
        num_examples // 10,
    )

    with torch.no_grad():
        trained_predictions = calibrated_linear(training_examples)
        trained_loss = loss_fn(trained_predictions, training_labels)

    assert trained_loss < initial_loss
    assert trained_loss < 0.02
